# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
import os.path
from typing import List, Union
import webbrowser
import numpy as np
import pandas as pd
# Engine imports

# GUI imports
from VeraGrid.Gui.gui_functions import get_list_model, get_logger_tree_model

from PySide6 import QtGui, QtCore
from VeraGrid.Gui.messages import *
from VeraGrid.Gui.RosetaExplorer.MainWindow import Ui_mainWindow, QMainWindow
from VeraGrid.Gui.RosetaExplorer.roseta_objects_model import RosetaObjectsModel, ObjectsModelOld
from VeraGrid.Gui.TreeModelViewer.TreeModelViewer import TreeModelViewerGUI
from VeraGrid.Gui.gui_functions import add_menu_entry
from VeraGrid.Gui.object_model import ObjectsModel
from VeraGrid.Gui.pandas_model import PandasModel
from VeraGridEngine.IO.cim.cgmes.cgmes_circuit import CgmesCircuit
from VeraGridEngine.IO.raw.devices.psse_circuit import PsseCircuit
from VeraGridEngine.IO.cim.db.db_handler import DbHandler
from VeraGridEngine.data_logger import DataLogger
from VeraGrid.Gui.python_console import PythonConsole


def clear_qt_layout(layout):
    """
    Remove all widgets from a layout object
    :param layout:
    """
    for i in reversed(range(layout.count())):
        layout.itemAt(i).widget().deleteLater()


########################################################################################################################
# Main Window
########################################################################################################################


class RosetaExplorerGUI(QMainWindow):
    """
    RosetaExplorerGUI
    """

    def __init__(self, parent=None, db_handler: Union[DbHandler, None] = None):
        """

        @param parent:
        """

        # create main window
        QMainWindow.__init__(self, parent)
        self.ui = Ui_mainWindow()
        self.ui.setupUi(self)
        self.title = 'Roseta Explorer'
        self.setWindowTitle(self.title)
        self.setAcceptDrops(True)

        # 1:4
        self.ui.main_splitter.setStretchFactor(0, 1)
        self.ui.main_splitter.setStretchFactor(1, 8)
        self.ui.lookup_db_splitter.setStretchFactor(0, 1)
        self.ui.lookup_db_splitter.setStretchFactor(1, 6)

        self.open_file_thread_object = None

        self.accepted_extensions = ['.zip', '.xml', '.raw']
        self.project_directory = ''
        self.lock_ui = True

        self.tree_navigation_windows: List[TreeModelViewerGUI] = list()

        self.circuit: Union[CgmesCircuit, PsseCircuit, None] = None

        self.logger: DataLogger = DataLogger()

        # Representation models
        self.classes_model = None
        self.properties_model = None
        self.properties_proxy_model = QtCore.QSortFilterProxyModel(self.ui.propertiesTableView)
        self.properties_proxy_model.setRecursiveFilteringEnabled(False)
        self.properties_proxy_model.setFilterKeyColumn(0)
        self.ui.propertiesTableView.setSortingEnabled(True)

        # DB
        self.db_handler: DbHandler = DbHandler(new_db=False) if db_handler is None else db_handler
        self.update_combo_boxes()

        # Console
        self.console = None
        self.create_console()

        # Connections --------------------------------------------------------------------------------------------------

        # triggered
        self.ui.actionSave_logs.triggered.connect(self.save_logs)

        self.ui.actionNavigation_tree.triggered.connect(self.launch_tree_view)

        self.ui.actionNew.triggered.connect(self.new_rosetta_explorer)

        self.ui.actionDocumentation.triggered.connect(self.show_docs)

        self.ui.actionAbout.triggered.connect(self.about)

        # clicked

        self.ui.filterButton.clicked.connect(self.filter_properties_table)
        self.ui.modelTypeLabel.mouseDoubleClickEvent = self.display_circuit

        # text changed

        self.ui.filterComboBox.currentTextChanged.connect(self.on_filter_combobox_changed)

        self.ui.availableDBsComboBox.currentTextChanged.connect(self.available_db_combo_box_changed)

        # lineedit enter

        self.ui.filterLineEdit.returnPressed.connect(self.filter_properties_table)

        # list clicks

        self.ui.clasesListView.clicked.connect(self.on_class_click)

        self.ui.dbTablesListView.clicked.connect(self.on_look_up_db_class_click)

        # context menu
        self.ui.propertiesTableView.customContextMenuRequested.connect(self.show_objects_context_menu)

        # Set context menu policy to CustomContextMenu
        self.ui.propertiesTableView.setContextMenuPolicy(QtGui.Qt.ContextMenuPolicy.CustomContextMenu)

        self.UNLOCK()
        self.update_combo_boxes()

    def LOCK(self, val: bool = True) -> None:
        """
        Lock the interface to prevent new simulation launches
        :param val:
        :return:
        """
        self.lock_ui = val
        self.ui.progress_frame.setVisible(self.lock_ui)
        QtGui.QGuiApplication.processEvents()

    def UNLOCK(self) -> None:
        """
        Unlock the interface
        """
        if not self.any_thread_running():
            self.LOCK(False)

    def create_console(self) -> None:
        """
        Create console
        """
        if self.console is not None:
            clear_qt_layout(self.ui.consoleLayout.layout())

        self.console = PythonConsole()

        # add the console widget to the user interface
        self.ui.consoleLayout.layout().addWidget(self.console)

        # push some variables to the console
        self.console.push_vars({"np": np,
                                "pd": pd,
                                'app': self,
                                'circuit': self.circuit})

    def update_combo_boxes(self):
        """
        Properly update te combo boxes
        """
        if isinstance(self.circuit, PsseCircuit):
            # we do this first so that the model changing does not trigger the on_combo_box_text_change function
            if self.db_handler.psse_lookup_db.last_file_opened in self.db_handler.psse_lookup_db.list_of_db_files:
                idx = self.db_handler.psse_lookup_db.list_of_db_files.index(
                    self.db_handler.psse_lookup_db.last_file_opened)
            else:
                idx = 0

            self.ui.availableDBsComboBox.setModel(get_list_model(self.db_handler.psse_lookup_db.list_of_db_files))
            self.ui.availableDBsComboBox.setCurrentIndex(idx)

        elif isinstance(self.circuit, CgmesCircuit):

            # we do this first so that the model changing does not trigger the on_combo_box_text_change function
            if self.db_handler.cgmes_lookup_db.last_file_opened in self.db_handler.cgmes_lookup_db.list_of_db_files:
                idx = self.db_handler.cgmes_lookup_db.list_of_db_files.index(
                    self.db_handler.cgmes_lookup_db.last_file_opened)
            else:
                idx = 0

            self.ui.availableDBsComboBox.setModel(get_list_model(self.db_handler.cgmes_lookup_db.list_of_db_files))
            self.ui.availableDBsComboBox.setCurrentIndex(idx)

        else:
            pass

    def get_process_threads(self):
        """
        Get all threads that has to do with processing
        :return: list of process threads
        """
        all_threads = [self.open_file_thread_object]
        return all_threads

    def get_all_threads(self):
        """
        Get all threads
        :return: list of all threads
        """
        all_threads = self.get_process_threads()
        return all_threads

    def any_thread_running(self):
        """
        Checks if any thread is running
        :return: True/False
        """
        val = False

        # this list cannot be created only once, because the None will be copied
        # instead of being a pointer to the future value like it would in a typed language
        all_threads = self.get_all_threads()

        for thr in all_threads:
            if thr is not None:
                if thr.isRunning():
                    return True
        return val

    def dragEnterEvent(self, event):
        """

        :param event:
        :return:
        """
        if event.mimeData().hasUrls:
            event.accept()
        else:
            event.ignore()

    def dragMoveEvent(self, event):
        """

        :param event:
        :return:
        """
        if event.mimeData().hasUrls:
            event.accept()
        else:
            event.ignore()

    def on_filter_combobox_changed(self):
        """
        Main property chooser for filtering table data
        """
        idx = self.ui.filterComboBox.currentIndex()

        if idx > -1:
            self.properties_proxy_model.setFilterKeyColumn(idx)

    def filter_properties_table(self):
        """
        Action of filtering the main table data
        """
        self.properties_proxy_model.setFilterRegularExpression(self.ui.filterLineEdit.text())

    def set_grid_model(self, circuit: Union[CgmesCircuit, PsseCircuit]):
        """
        Set a grid Model
        :param circuit: circuit model passed
        :param
        """
        if type(circuit) in [CgmesCircuit, PsseCircuit]:
            self.circuit = circuit

            self.display_circuit(trim_empty=True)

    def display_circuit(self, trim_empty=False):
        """

        :param trim_empty:
        :return:
        """
        if self.circuit is not None:
            classes = [prop.property_name for prop in self.circuit.get_class_properties()]
            classes.sort()

            if trim_empty:
                classes = [cl for cl in classes if len(self.circuit.get_objects_list(elm_type=cl)) > 0]

            self.classes_model = get_list_model(classes)

            self.ui.clasesListView.setModel(self.classes_model)

            self.set_model_label()
            self.update_combo_boxes()
            self.available_db_combo_box_changed()

            if self.console:
                self.console.push_vars({'app': self,
                                        'circuit': self.circuit})

    def set_model_label(self):
        """
        Set the model label based on the model inner type
        """
        if isinstance(self.circuit, PsseCircuit):
            self.ui.modelTypeLabel.setText("PSSe")
            self.setWindowTitle(self.title + ': PSSe')

        elif isinstance(self.circuit, CgmesCircuit):
            self.ui.modelTypeLabel.setText("CGMES")
            self.setWindowTitle(self.title + ': CGMES')

        else:
            raise Exception('Unknown circuit type :(')

    def set_logger(self, logger: DataLogger, message=True):
        """
        Set the main logger
        :param logger: DataLogger object
        :param message: Open message if relevant
        """
        self.logger = logger  # assign the logger pointer
        logger_model = get_logger_tree_model(logger)  # crete tree model
        self.ui.loggerTreeView.setModel(logger_model)  # display

        if logger.size():
            error_msg(logger.get_message(), 'Logger')

    def on_class_click(self):
        """
        On click on the classes list...
        """
        # get the class type
        elm_type = self.ui.clasesListView.selectedIndexes()[0].data(role=QtCore.Qt.ItemDataRole.DisplayRole)

        # get the objects list of said class depending on the circuit type
        if isinstance(self.circuit, PsseCircuit):
            objects = self.circuit.get_objects_list(elm_type)

        elif isinstance(self.circuit, CgmesCircuit):
            objects = self.circuit.get_objects_list(elm_type)

        else:
            raise Exception('Unknown circuit type :(')

        # represent the model
        if len(objects) > 0:

            editable_headers = {p.property_name: p for p in objects[0].get_properties()}
            headers = [p.property_name for p in objects[0].get_properties()]

        else:

            editable_headers = dict()
            headers = list()

        self.ui.filterComboBox.setModel(get_list_model(headers))

        self.properties_model = RosetaObjectsModel(objects=objects,
                                                   editable_headers=editable_headers,
                                                   parent=self.ui.propertiesTableView,
                                                   editable=True,
                                                   non_editable_attributes=[],
                                                   dictionary_of_lists={})

        self.properties_proxy_model.setSourceModel(self.properties_model)
        self.ui.propertiesTableView.setModel(self.properties_proxy_model)

    def save_logs(self):
        """
        Save logger content
        """
        if self.logger.size():

            file, selected_filter = QtWidgets.QFileDialog.getSaveFileName(self, "Export logs", '',
                                                                          filter="Excel files (*.xlsx)")

            if file != '':
                if 'xlsx' in selected_filter:
                    f = file
                    if not f.endswith('.xlsx'):
                        f += '.xlsx'
                    self.logger.to_xlsx(f)

                else:
                    error_msg(file[0] + ' is not valid :(')
        else:
            warning_msg('There no logs :)')

    def launch_tree_view(self):
        """
        Launch tree view of the model
        """
        if self.circuit is not None:

            if isinstance(self.circuit, CgmesCircuit):
                window = TreeModelViewerGUI()
                self.tree_navigation_windows.append(window)
                h = 740
                window.resize(int(1.61 * h), h)  # golden ratio :)

                window.set_circuit(self.circuit)

                window.show()
            else:
                info_msg("Only CGMES models are available for the tree view :/")
        else:
            warning_msg("There is not model :/")

    def new_rosetta_explorer(self, model: Union[PsseCircuit, CgmesCircuit, None] = None,
                             logger: Union[DataLogger, None] = None):
        """
        New roseta window
        :param model: (optional) the model to load
        :param logger: (optional) DataLogger to display
        """
        window = RosetaExplorerGUI(db_handler=self.db_handler)
        self.tree_navigation_windows.append(window)
        h = 740
        window.resize(int(1.61 * h), h)  # golden ratio :)

        if model is not None:
            window.set_grid_model(model)

        if logger is not None:
            window.set_logger(logger)

        window.show()

    def available_db_combo_box_changed(self):
        """
        Read PSSe lookup DB
        """
        name = self.ui.availableDBsComboBox.currentText()

        if isinstance(self.circuit, PsseCircuit):
            self.db_handler.psse_lookup_db.read_db_file(name)
            self.ui.dbTablesListView.setModel(get_list_model(self.db_handler.psse_lookup_db.get_structures_names()))
            self.on_look_up_db_class_click()

        elif isinstance(self.circuit, CgmesCircuit):
            self.db_handler.cgmes_lookup_db.read_db_file(name)
            self.ui.dbTablesListView.setModel(get_list_model(self.db_handler.cgmes_lookup_db.get_structures_names()))
            self.on_look_up_db_class_click()

    def on_look_up_db_class_click(self):
        """

        :return:
        """

        if len(self.ui.dbTablesListView.selectedIndexes()) > 0:
            elm_type = self.ui.dbTablesListView.selectedIndexes()[0].data(role=QtCore.Qt.ItemDataRole.DisplayRole)

            if isinstance(self.circuit, PsseCircuit):
                db = self.db_handler.psse_lookup_db
                df = getattr(db, elm_type)
                mdl = PandasModel(data=df)

            elif isinstance(self.circuit, CgmesCircuit):
                objects = self.db_handler.cgmes_lookup_db.circuit.elements_by_type.get(elm_type, [])
                if len(objects) > 0:
                    editable_headers = {p.property_name: p for p in objects[0].get_properties()}
                    headers = [p.property_name for p in objects[0].get_properties()]

                else:
                    editable_headers = dict()
                    headers = list()

                mdl = ObjectsModelOld(objects=objects,
                                      editable_headers=editable_headers,
                                      parent=self.ui.propertiesTableView,
                                      editable=True,
                                      dictionary_of_lists={})
            else:
                return

            # set the model
            self.ui.dbTableView.setModel(mdl)
        else:
            self.ui.dbTableView.setModel(None)

    def show_objects_context_menu(self, pos: QtCore.QPoint):
        """
        Show diagrams list view context menu
        :param pos: Relative click position
        """
        context_menu = QtWidgets.QMenu(parent=self.ui.propertiesTableView)

        add_menu_entry(menu=context_menu,
                       text="Copy",
                       icon_path=":/Icons/icons/copy.svg",
                       function_ptr=self.copy_table_to_clipboard)

    def copy_table_to_clipboard(self) -> None:
        """
        Copy objects model to the clipboard
        """
        mdl = self.ui.propertiesTableView.model()

        if isinstance(mdl, QtCore.QSortFilterProxyModel):
            mdl = mdl.sourceModel()

        if isinstance(mdl, RosetaObjectsModel):
            mdl.copy_to_clipboard()
            info_msg("Copied table to clipboard!")
        elif isinstance(mdl, ObjectsModel):
            mdl.copy_to_clipboard()
            info_msg("Copied table to clipboard!")
        elif isinstance(mdl, PandasModel):
            mdl.copy_to_clipboard()
            info_msg("Copied table to clipboard!")

    @staticmethod
    def show_docs():
        """
        Show the docs
        """
        this_py_file = os.path.realpath(__file__)
        current_folder_path, current_folder_name = os.path.split(this_py_file)
        index_path = os.path.join(current_folder_path, '..', '..', 'docs', 'html', 'index.html')
        if os.path.exists(index_path):
            webbrowser.open(index_path, new=2)
        else:
            error_msg("The documentation could not be found under " + index_path)

    def about(self):
        """
        Show the "about" dialogue
        :return:
        """
        pass



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
"""
Script to update correctly the main GUI (.py) file from the Qt design (.ui) file
"""
import os
from VeraGrid.Gui.update_gui_common import convert_resource_file, convert_ui_file

if __name__ == '__main__':
    # pyrcc5 icons.qrc -o icons_rc.py
    # pyuic5 -x MainWindow.ui -o MainWindow.py

    rcc_cmd = 'pyside6-rcc'
    uic_cmd = 'pyside6-uic'

    if os.name == 'nt':
        rcc_cmd += '.exe'
        uic_cmd += '.exe'

    convert_resource_file(source='icons.qrc', rcc_cmd=rcc_cmd)
    for f in ['MainWindow.ui']:
        convert_ui_file(source=f, uic_cmd=uic_cmd)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

import numpy as np
from typing import List

from PySide6 import QtCore, QtWidgets
from VeraGrid.Gui.SolarPowerWizard.solar_power_wizzard import get_pv_lib_weather_df


class WindFarmWizard(QtWidgets.QDialog):
    """
    New solar photovoltaic wizard window
    """

    def __init__(self, time_array: List[str], peak_power: float, latitude: float, longitude: float,
                 gen_name='', bus_name='',
                 title='Wind farm wizard'):
        """

        :param time_array: array of time values
        :param peak_power: generator peak power in MW
        :param latitude: latitude (float)
        :param longitude: longitude (float)
        :param title: Window title
        """
        QtWidgets.QDialog.__init__(self)
        self.setObjectName("self")
        self.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.NoContextMenu)
        self.main_layout = QtWidgets.QVBoxLayout(self)

        self.is_accepted: bool = False
        self.selected_indices: List[int] = list()

        self.label_gen = QtWidgets.QLabel()
        self.label_gen.setText("Generator {}".format(gen_name))

        self.label_bus = QtWidgets.QLabel()
        self.label_bus.setText("Bus: {}".format(bus_name))

        self.label_peak = QtWidgets.QLabel()
        self.label_peak.setText("peak power {} MW".format(peak_power))

        self.lat_label = QtWidgets.QLabel()
        self.lat_label.setText("Latitude {} deg".format(latitude))

        self.lon_label = QtWidgets.QLabel()
        self.lon_label.setText("Longitude {} deg".format(longitude))

        self.peak_power = peak_power
        self.latitude = latitude
        self.longitude = longitude
        self.time_array = time_array
        self.P = np.zeros(len(time_array))

        # accept button
        self.accept_btn = QtWidgets.QPushButton()
        self.accept_btn.setText('Accept')
        self.accept_btn.clicked.connect(self.accept_click)

        # add all to the GUI
        self.main_layout.addWidget(self.label_gen)
        self.main_layout.addWidget(self.label_bus)
        self.main_layout.addWidget(self.label_peak)
        self.main_layout.addWidget(self.lat_label)
        self.main_layout.addWidget(self.lon_label)
        self.main_layout.addWidget(self.accept_btn)

        self.setLayout(self.main_layout)

        self.setWindowTitle(title)

        h = 260
        self.resize(h, int(0.8 * h))

    def compute(self, wind_speed):
        pass

    def accept_click(self):
        """
        Accept and close
        """
        ok, df = get_pv_lib_weather_df(time_array=self.time_array,
                                       latitude=self.latitude,
                                       longitude=self.longitude,
                                       peak_power=self.peak_power)

        if ok:
            self.P = df['P'].values / 1e6  # Power in MW

        self.is_accepted = ok
        self.accept()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'profiles_from_data_gui.ui'
##
## Created by: Qt User Interface Compiler version 6.7.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QCheckBox, QComboBox, QDialog,
    QDoubleSpinBox, QFrame, QGridLayout, QHBoxLayout,
    QHeaderView, QLabel, QListView, QPushButton,
    QSizePolicy, QSlider, QSpacerItem, QSplitter,
    QTableView, QVBoxLayout, QWidget)
from .icons_rc import *

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        if not Dialog.objectName():
            Dialog.setObjectName(u"Dialog")
        Dialog.resize(1036, 551)
        self.verticalLayout = QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.splitter = QSplitter(Dialog)
        self.splitter.setObjectName(u"splitter")
        self.splitter.setOrientation(Qt.Orientation.Horizontal)
        self.frame_4 = QFrame(self.splitter)
        self.frame_4.setObjectName(u"frame_4")
        self.frame_4.setFrameShape(QFrame.Shape.NoFrame)
        self.frame_4.setFrameShadow(QFrame.Shadow.Raised)
        self.verticalLayout_5 = QVBoxLayout(self.frame_4)
        self.verticalLayout_5.setObjectName(u"verticalLayout_5")
        self.verticalLayout_5.setContentsMargins(-1, 0, -1, 0)
        self.frame = QFrame(self.frame_4)
        self.frame.setObjectName(u"frame")
        self.frame.setMaximumSize(QSize(16777215, 40))
        self.frame.setFrameShape(QFrame.Shape.NoFrame)
        self.frame.setFrameShadow(QFrame.Shadow.Raised)
        self.horizontalLayout_5 = QHBoxLayout(self.frame)
        self.horizontalLayout_5.setObjectName(u"horizontalLayout_5")
        self.horizontalLayout_5.setContentsMargins(0, 4, 0, 0)
        self.open_button = QPushButton(self.frame)
        self.open_button.setObjectName(u"open_button")
        icon = QIcon()
        icon.addFile(u":/Icons/icons/import_profiles.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.open_button.setIcon(icon)

        self.horizontalLayout_5.addWidget(self.open_button)

        self.plotButton = QPushButton(self.frame)
        self.plotButton.setObjectName(u"plotButton")
        icon1 = QIcon()
        icon1.addFile(u":/Icons/icons/plot.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.plotButton.setIcon(icon1)

        self.horizontalLayout_5.addWidget(self.plotButton)

        self.horizontalSpacer_2 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_5.addItem(self.horizontalSpacer_2)

        self.label_5 = QLabel(self.frame)
        self.label_5.setObjectName(u"label_5")

        self.horizontalLayout_5.addWidget(self.label_5)

        self.units_combobox = QComboBox(self.frame)
        self.units_combobox.setObjectName(u"units_combobox")

        self.horizontalLayout_5.addWidget(self.units_combobox)


        self.verticalLayout_5.addWidget(self.frame)

        self.sources_list = QListView(self.frame_4)
        self.sources_list.setObjectName(u"sources_list")

        self.verticalLayout_5.addWidget(self.sources_list)

        self.frame_2 = QFrame(self.frame_4)
        self.frame_2.setObjectName(u"frame_2")
        self.frame_2.setFrameShape(QFrame.Shape.NoFrame)
        self.frame_2.setFrameShadow(QFrame.Shadow.Raised)
        self.horizontalLayout_3 = QHBoxLayout(self.frame_2)
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.nameTransformationComboBox = QComboBox(self.frame_2)
        self.nameTransformationComboBox.setObjectName(u"nameTransformationComboBox")
        self.nameTransformationComboBox.setMinimumSize(QSize(200, 0))

        self.horizontalLayout_3.addWidget(self.nameTransformationComboBox)

        self.transformNamesPushButton = QPushButton(self.frame_2)
        self.transformNamesPushButton.setObjectName(u"transformNamesPushButton")
        icon2 = QIcon()
        icon2.addFile(u":/Icons/icons/transform.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.transformNamesPushButton.setIcon(icon2)

        self.horizontalLayout_3.addWidget(self.transformNamesPushButton)

        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_3.addItem(self.horizontalSpacer)


        self.verticalLayout_5.addWidget(self.frame_2)

        self.splitter.addWidget(self.frame_4)
        self.frame_6 = QFrame(self.splitter)
        self.frame_6.setObjectName(u"frame_6")
        self.frame_6.setFrameShape(QFrame.Shape.NoFrame)
        self.frame_6.setFrameShadow(QFrame.Shadow.Raised)
        self.verticalLayout_3 = QVBoxLayout(self.frame_6)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.verticalLayout_3.setContentsMargins(0, 0, 9, -1)
        self.frame_10 = QFrame(self.frame_6)
        self.frame_10.setObjectName(u"frame_10")
        self.frame_10.setMaximumSize(QSize(16777215, 40))
        self.frame_10.setFrameShape(QFrame.Shape.NoFrame)
        self.frame_10.setFrameShadow(QFrame.Shadow.Raised)
        self.horizontalLayout_7 = QHBoxLayout(self.frame_10)
        self.horizontalLayout_7.setObjectName(u"horizontalLayout_7")
        self.horizontalLayout_7.setContentsMargins(0, 4, 0, 0)
        self.autolink_button = QPushButton(self.frame_10)
        self.autolink_button.setObjectName(u"autolink_button")
        icon3 = QIcon()
        icon3.addFile(u":/Icons/icons/auto-link.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.autolink_button.setIcon(icon3)

        self.horizontalLayout_7.addWidget(self.autolink_button)

        self.autolink_slider = QSlider(self.frame_10)
        self.autolink_slider.setObjectName(u"autolink_slider")
        self.autolink_slider.setMinimum(1)
        self.autolink_slider.setMaximum(100)
        self.autolink_slider.setValue(60)
        self.autolink_slider.setOrientation(Qt.Orientation.Horizontal)

        self.horizontalLayout_7.addWidget(self.autolink_slider)

        self.rnd_link_pushButton = QPushButton(self.frame_10)
        self.rnd_link_pushButton.setObjectName(u"rnd_link_pushButton")
        icon4 = QIcon()
        icon4.addFile(u":/Icons/icons/auto-link-rnd.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.rnd_link_pushButton.setIcon(icon4)

        self.horizontalLayout_7.addWidget(self.rnd_link_pushButton)

        self.assign_to_all_pushButton = QPushButton(self.frame_10)
        self.assign_to_all_pushButton.setObjectName(u"assign_to_all_pushButton")
        icon5 = QIcon()
        icon5.addFile(u":/Icons/icons/link-to-all.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.assign_to_all_pushButton.setIcon(icon5)

        self.horizontalLayout_7.addWidget(self.assign_to_all_pushButton)

        self.assign_to_selection_pushButton = QPushButton(self.frame_10)
        self.assign_to_selection_pushButton.setObjectName(u"assign_to_selection_pushButton")
        icon6 = QIcon()
        icon6.addFile(u":/Icons/icons/link-to-selection.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.assign_to_selection_pushButton.setIcon(icon6)

        self.horizontalLayout_7.addWidget(self.assign_to_selection_pushButton)

        self.horizontalSpacer_4 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_7.addItem(self.horizontalSpacer_4)


        self.verticalLayout_3.addWidget(self.frame_10)

        self.assignation_table = QTableView(self.frame_6)
        self.assignation_table.setObjectName(u"assignation_table")

        self.verticalLayout_3.addWidget(self.assignation_table)

        self.frame_9 = QFrame(self.frame_6)
        self.frame_9.setObjectName(u"frame_9")
        self.frame_9.setFrameShape(QFrame.Shape.NoFrame)
        self.frame_9.setFrameShadow(QFrame.Shadow.Raised)
        self.gridLayout = QGridLayout(self.frame_9)
        self.gridLayout.setObjectName(u"gridLayout")
        self.gridLayout.setContentsMargins(0, -1, 0, 0)
        self.set_multiplier_button = QPushButton(self.frame_9)
        self.set_multiplier_button.setObjectName(u"set_multiplier_button")
        icon7 = QIcon()
        icon7.addFile(u":/Icons/icons/multiply.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.set_multiplier_button.setIcon(icon7)

        self.gridLayout.addWidget(self.set_multiplier_button, 0, 0, 1, 1)

        self.setUnassignedToZeroCheckBox = QCheckBox(self.frame_9)
        self.setUnassignedToZeroCheckBox.setObjectName(u"setUnassignedToZeroCheckBox")
        self.setUnassignedToZeroCheckBox.setChecked(True)

        self.gridLayout.addWidget(self.setUnassignedToZeroCheckBox, 0, 5, 1, 1)

        self.horizontalSpacer_3 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.gridLayout.addItem(self.horizontalSpacer_3, 0, 3, 1, 1)

        self.normalized_checkBox = QCheckBox(self.frame_9)
        self.normalized_checkBox.setObjectName(u"normalized_checkBox")

        self.gridLayout.addWidget(self.normalized_checkBox, 0, 6, 1, 1)

        self.multSpinBox = QDoubleSpinBox(self.frame_9)
        self.multSpinBox.setObjectName(u"multSpinBox")
        self.multSpinBox.setMinimum(-99999.000000000000000)
        self.multSpinBox.setMaximum(99999.000000000000000)

        self.gridLayout.addWidget(self.multSpinBox, 0, 1, 1, 1)

        self.doit_button = QPushButton(self.frame_9)
        self.doit_button.setObjectName(u"doit_button")
        icon8 = QIcon()
        icon8.addFile(u":/Icons/icons/gear.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.doit_button.setIcon(icon8)

        self.gridLayout.addWidget(self.doit_button, 0, 7, 1, 1)

        self.clear_selection_button = QPushButton(self.frame_9)
        self.clear_selection_button.setObjectName(u"clear_selection_button")
        icon9 = QIcon()
        icon9.addFile(u":/Icons/icons/clear_runs.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.clear_selection_button.setIcon(icon9)

        self.gridLayout.addWidget(self.clear_selection_button, 0, 4, 1, 1)


        self.verticalLayout_3.addWidget(self.frame_9)

        self.splitter.addWidget(self.frame_6)

        self.verticalLayout.addWidget(self.splitter)


        self.retranslateUi(Dialog)

        QMetaObject.connectSlotsByName(Dialog)
    # setupUi

    def retranslateUi(self, Dialog):
        Dialog.setWindowTitle(QCoreApplication.translate("Dialog", u"Dialog", None))
#if QT_CONFIG(tooltip)
        self.open_button.setToolTip(QCoreApplication.translate("Dialog", u"Import file", None))
#endif // QT_CONFIG(tooltip)
        self.open_button.setText("")
#if QT_CONFIG(tooltip)
        self.plotButton.setToolTip(QCoreApplication.translate("Dialog", u"Plot selected profile", None))
#endif // QT_CONFIG(tooltip)
        self.plotButton.setText("")
        self.label_5.setText(QCoreApplication.translate("Dialog", u"Units", None))
#if QT_CONFIG(tooltip)
        self.transformNamesPushButton.setToolTip(QCoreApplication.translate("Dialog", u"Transform the input profile names", None))
#endif // QT_CONFIG(tooltip)
        self.transformNamesPushButton.setText("")
#if QT_CONFIG(tooltip)
        self.autolink_button.setToolTip(QCoreApplication.translate("Dialog", u"Automatic link profiles to objects based on their name", None))
#endif // QT_CONFIG(tooltip)
        self.autolink_button.setText("")
#if QT_CONFIG(tooltip)
        self.autolink_slider.setToolTip(QCoreApplication.translate("Dialog", u"Auto-link string simmilarity", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(tooltip)
        self.rnd_link_pushButton.setToolTip(QCoreApplication.translate("Dialog", u"Random-link profiles and objects", None))
#endif // QT_CONFIG(tooltip)
        self.rnd_link_pushButton.setText("")
#if QT_CONFIG(tooltip)
        self.assign_to_all_pushButton.setToolTip(QCoreApplication.translate("Dialog", u"Assign profile to all objects", None))
#endif // QT_CONFIG(tooltip)
        self.assign_to_all_pushButton.setText("")
#if QT_CONFIG(tooltip)
        self.assign_to_selection_pushButton.setToolTip(QCoreApplication.translate("Dialog", u"Assign profile to object selection", None))
#endif // QT_CONFIG(tooltip)
        self.assign_to_selection_pushButton.setText("")
#if QT_CONFIG(tooltip)
        self.set_multiplier_button.setToolTip(QCoreApplication.translate("Dialog", u"Set multiplier", None))
#endif // QT_CONFIG(tooltip)
        self.set_multiplier_button.setText("")
#if QT_CONFIG(statustip)
        self.setUnassignedToZeroCheckBox.setStatusTip(QCoreApplication.translate("Dialog", u"The profiles for the unassigned objects are set to zero, otherwise they are not set and they remain the default value form the snapshot", None))
#endif // QT_CONFIG(statustip)
        self.setUnassignedToZeroCheckBox.setText(QCoreApplication.translate("Dialog", u"Set unassigned to zero", None))
#if QT_CONFIG(tooltip)
        self.normalized_checkBox.setToolTip(QCoreApplication.translate("Dialog", u"Check if you want the profiles to be normalized on the base object property", None))
#endif // QT_CONFIG(tooltip)
        self.normalized_checkBox.setText(QCoreApplication.translate("Dialog", u"normalized", None))
#if QT_CONFIG(tooltip)
        self.doit_button.setToolTip(QCoreApplication.translate("Dialog", u"Do it!", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(statustip)
        self.doit_button.setStatusTip("")
#endif // QT_CONFIG(statustip)
        self.doit_button.setText(QCoreApplication.translate("Dialog", u"Accept", None))
#if QT_CONFIG(tooltip)
        self.clear_selection_button.setToolTip(QCoreApplication.translate("Dialog", u"Clear selection", None))
#endif // QT_CONFIG(tooltip)
        self.clear_selection_button.setText("")
    # retranslateUi




# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
from __future__ import annotations
import numpy as np
from typing import List, TYPE_CHECKING
from PySide6.QtCore import Qt, QPoint, QPointF
from PySide6.QtGui import QPen, QCursor, QColor
from PySide6.QtWidgets import QGraphicsItem, QGraphicsEllipseItem, QGraphicsRectItem, QMenu, QGraphicsSceneMouseEvent

from VeraGrid.Gui.Diagrams.Editors.transformer3w_editor import Transformer3WEditor
from VeraGrid.Gui.Diagrams.generic_graphics import ACTIVE, DEACTIVATED, GenericDiagramWidget
from VeraGrid.Gui.Diagrams.SchematicWidget.terminal_item import RoundTerminalItem
from VeraGrid.Gui.Diagrams.SchematicWidget.Branches.winding_graphics import WindingGraphicItem
from VeraGridEngine.Devices.Branches.transformer3w import Transformer3W
from VeraGridEngine.Devices.Substation.bus import Bus
from VeraGrid.Gui.messages import yes_no_question
from VeraGridEngine.enumerations import DeviceType
from VeraGrid.Gui.gui_functions import add_menu_entry

if TYPE_CHECKING:
    # Only imports the below statements during type checking
    from VeraGrid.Gui.Diagrams.SchematicWidget.schematic_widget import SchematicWidget


class Transformer3WGraphicItem(GenericDiagramWidget, QGraphicsRectItem):
    """
      Represents a block in the diagram
      Has an x and y and width and height
      width and height can only be adjusted with a tip in the lower right corner.

      - in and output ports
      - parameters
      - description
    """

    def __init__(self, editor: SchematicWidget,
                 elm: Transformer3W,
                 pos: QPoint = None,
                 parent=None,
                 index=0,
                 draw_labels: bool = True):
        """

        :param editor: GridEditor object
        :param elm: Transformer3W object
        :param pos: position
        :param parent:
        :param index:
        :param draw_labels:
        """
        self.h = 100
        self.w = 100
        GenericDiagramWidget.__init__(self, parent=parent, api_object=elm, editor=editor, draw_labels=True)
        QGraphicsRectItem.__init__(self, 0.0, 0.0, self.w, self.h, parent=parent)
        self.n_windings = 3

        self.draw_labels = draw_labels

        # color
        self.pen_width = 4
        if self.api_object is not None:
            if self.api_object.active:
                self.color = ACTIVE['color']
                self.style = ACTIVE['style']
            else:
                self.color = DEACTIVATED['color']
                self.style = DEACTIVATED['style']
        else:
            self.color = ACTIVE['color']
            self.style = ACTIVE['style']

        self.setPen(QPen(Qt.GlobalColor.transparent, self.pen_width, self.style))
        self.setBrush(Qt.GlobalColor.transparent)
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsMovable)
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsSelectable)
        self.setCursor(QCursor(Qt.CursorShape.PointingHandCursor))

        # index
        self.index = index

        if pos is not None:
            self.setPos(pos)

        # windings
        winding_diameter = 40
        radius_large = winding_diameter * 0.5
        radius_small = 5
        angle_0 = -90.0
        d_angle = 360.0 / self.n_windings
        center = QPointF(self.w / 2.0, self.h / 2.0)

        self.winding_circles: List[QGraphicsEllipseItem] = list()
        self.terminals: List[RoundTerminalItem] = list()
        self.connection_lines: List[WindingGraphicItem | None] = list()

        pen = QPen(self.color, self.pen_width, self.style)

        for i in range(self.n_windings):
            angles_deg = angle_0 + d_angle * i
            angle_rad = np.deg2rad(angles_deg)

            # Direction vector
            dx = np.cos(angle_rad)
            dy = np.sin(angle_rad)

            winding_center_x = center.x() + dx * radius_large - radius_large
            winding_center_y = center.y() + dy * radius_large - radius_large

            # create objects
            winding_circle = QGraphicsEllipseItem(parent=self)
            winding_circle.setRect(winding_center_x, winding_center_y, winding_diameter, winding_diameter)

            # Compute offset to make small circle tangent and centered on the same line
            tangent_center_distance = winding_diameter + radius_small
            terminal_center_x = center.x() + dx * tangent_center_distance - radius_small
            terminal_center_y = center.y() + dy * tangent_center_distance - radius_small

            terminal = RoundTerminalItem(name=f"t{i + 1}", parent=self, editor=self.editor, h=10, w=10)
            terminal.setPos(terminal_center_x, terminal_center_y)

            # set objects style
            winding_circle.setPen(pen)
            terminal.setPen(pen)

            self.winding_circles.append(winding_circle)
            self.terminals.append(terminal)
            self.connection_lines.append(None)

        self.big_marker = None

        # other actions
        self.set_winding_tool_tips()

    @property
    def api_object(self) -> Transformer3W:
        return self._api_object

    @property
    def editor(self) -> SchematicWidget:
        return self._editor

    def get_associated_widgets(self) -> List[WindingGraphicItem]:
        """

        :return:
        """
        return self.connection_lines

    def get_extra_graphics(self):
        """
        Get a list of all QGraphicsItem that are not GenericDiagramWidget elements associated with this widget.
        :return:
        """
        return self.winding_circles + self.terminals

    def recolour_mode(self):
        """
        Change the colour according to the system theme
        """
        if self.api_object is not None:
            if self.api_object.active:
                self.color = ACTIVE['color']
                self.style = ACTIVE['style']
            else:
                self.color = DEACTIVATED['color']
                self.style = DEACTIVATED['style']
        else:
            self.color = ACTIVE['color']
            self.style = ACTIVE['style']

        pen = QPen(self.color, self.pen_width, self.style)
        for i in range(self.n_windings):
            self.winding_circles[i].setPen(pen)
            self.terminals[i].setPen(pen)

            if self.connection_lines[i] is not None:
                self.connection_lines[i].recolour_mode()

    def set_enable(self, val=True):
        """
        Set the enable value, graphically and in the API
        @param val:
        @return:
        """
        self.api_object.active = val
        self.api_object.winding1.active = val
        self.api_object.winding2.active = val
        self.api_object.winding3.active = val

        self.recolour_mode()

    def set_winding_tool_tips(self) -> None:
        """
        Set
        :return:
        """
        if self.api_object is not None:
            self.winding_circles[0].setToolTip("Winding 1: {0} KV".format(self.api_object.V1))
            self.winding_circles[1].setToolTip("Winding 2: {0} KV".format(self.api_object.V2))
            self.winding_circles[2].setToolTip("Winding 3: {0} KV".format(self.api_object.V3))

    def set_label(self, val: str):
        """
        Set the label content
        :param val:
        :return:
        """
        # this function is just for compatibility
        pass

    def mouseMoveEvent(self, event: 'QGraphicsSceneMouseEvent'):
        """
        On mouse move of this object...
        Args:
            event: QGraphicsSceneMouseEvent inherited
        """
        super().mouseMoveEvent(event)

        self.editor.update_diagram_element(device=self.api_object,
                                           x=self.pos().x(),
                                           y=self.pos().y(),
                                           w=self.w,
                                           h=self.h,
                                           r=self.rotation(),
                                           graphic_object=self)

        # self.update_conn()

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent):
        """
        mouse press: display the editor
        :param event:
        :return:
        """
        if self.api_object is not None:
            self.editor.set_editor_model(api_object=self.api_object)

    def mouseDoubleClickEvent(self, event):
        """
        On double-click, edit
        :param event:
        :return:
        """

        if self.api_object is not None:
            if self.api_object.device_type in [DeviceType.Transformer3WDevice]:
                # trigger the editor
                self.edit()

    def contextMenuEvent(self, event):
        """
        Show context menu
        @param event:
        @return:
        """
        if self.api_object is not None:
            menu = QMenu()
            menu.addSection("3w-Transformer")

            add_menu_entry(menu=menu,
                           text="Active",
                           function_ptr=self.enable_disable_toggle,
                           checkeable=True,
                           checked_value=self.api_object.active)

            menu.addSeparator()

            add_menu_entry(menu=menu,
                           text="Edit",
                           function_ptr=self.edit,
                           icon_path=":/Icons/icons/edit.svg")

            add_menu_entry(menu=menu,
                           text="Delete",
                           function_ptr=self.delete,
                           icon_path=":/Icons/icons/delete_schematic.svg")

            menu.exec_(event.screenPos())
        else:
            pass

    def enable_disable_toggle(self):
        """

        @return:
        """
        if self.api_object is not None:
            if self.api_object.active:
                self.set_enable(False)
            else:
                self.set_enable(True)

            if self._editor.circuit.get_time_number() > 0:
                ok = yes_no_question('Do you want to update the time series active status accordingly?',
                                     'Update time series active status')

                if ok:
                    # change the bus state (time series)
                    self._editor.set_active_status_to_profile(self.api_object, override_question=True)
                    self._editor.set_active_status_to_profile(self.api_object.winding1, override_question=True)
                    self._editor.set_active_status_to_profile(self.api_object.winding2, override_question=True)
                    self._editor.set_active_status_to_profile(self.api_object.winding3, override_question=True)

    def add_big_marker(self, color=Qt.GlobalColor.red, tool_tip_text=""):
        """
        Add a big marker to the bus
        :param color: Qt Color ot the marker
        :param tool_tip_text: tool tip text to display
        :return:
        """
        if self.big_marker is None:
            self.big_marker = QGraphicsEllipseItem(0, 0, 180, 180, parent=self)
            self.big_marker.setBrush(color)
            self.big_marker.setOpacity(0.5)
            self.big_marker.setToolTip(tool_tip_text)

    def delete_big_marker(self):
        """
        Delete the big marker
        """
        if self.big_marker is not None:
            self.editor._remove_from_scene(self.big_marker)
            self.big_marker = None

    def change_size(self, w, h):
        """

        :param w:
        :param h:
        """
        # Keep for compatibility

        self.editor.update_diagram_element(device=self.api_object,
                                           x=self.pos().x(),
                                           y=self.pos().y(),
                                           w=w,
                                           h=h,
                                           r=self.rotation(),
                                           graphic_object=self)

    def set_position(self, x: float, y: float) -> None:
        """
        Set the bus x, y position
        :param x: x in pixels
        :param y: y in pixels
        """
        x = 0 if np.isnan(x) else int(x)
        y = 0 if np.isnan(y) else int(y)
        self.setPos(QPoint(int(x), int(y)))

    def get_connection_winding(self, from_port: RoundTerminalItem, to_port: RoundTerminalItem):
        """
        Find the winding between the terminals
        :param from_port: "from" terminal [TerminalItem]
        :param to_port: "to" terminal [TerminalItem]
        """
        for i, t in enumerate(self.terminals):
            if t in [from_port, to_port]:
                return i

        raise Exception("Unknown winding")

    def set_connection(self, i: int, bus: Bus, conn: WindingGraphicItem, set_voltage: bool = True):
        """
        Create the connection with a bus
        :param i: winding index 0-2
        :param bus: Bus object to connect to
        :param conn: Connection graphical object [LineGraphicItem]
        :param set_voltage: Set voltage in the object (Transformer3W and Windings)
        """
        if i == 0:
            self.api_object.bus1 = bus
            if set_voltage:
                self.api_object.V1 = bus.Vnom
            self.connection_lines[0] = conn
            self.terminals[0].setZValue(-1)
            conn.api_object = self.api_object.winding1
            conn.winding_number = 0

        elif i == 1:
            self.api_object.bus2 = bus
            if set_voltage:
                self.api_object.V2 = bus.Vnom
            self.connection_lines[1] = conn
            self.terminals[1].setZValue(-1)
            conn.api_object = self.api_object.winding2
            conn.winding_number = 1

        elif i == 2:
            self.api_object.bus3 = bus
            if set_voltage:
                self.api_object.V3 = bus.Vnom
            self.connection_lines[2] = conn
            self.terminals[2].setZValue(-1)
            conn.api_object = self.api_object.winding3
            conn.winding_number = 2
        else:
            raise Exception('Unsupported winding index {}'.format(i))

        # set the reverse lookup
        conn.parent_tr3_graphics_item = self

        # update the connection placement
        # self.update_conn()
        self.mousePressEvent(None)

    def remove_winding(self, i: int):
        """
        Remove winding by index
        :param i: winding index [0, 1, 2]
        """
        if i == 0:
            self.api_object.bus1 = None
            self.api_object.V1 = 0
            self.connection_lines[0] = None
            self.api_object.winding1.bus_to = None
            self.terminals[0].setZValue(-1)

        elif i == 1:
            self.api_object.bus2 = None
            self.api_object.V2 = 0
            self.connection_lines[1] = None
            self.api_object.winding2.bus_to = None
            self.terminals[1].setZValue(-1)

        elif i == 2:
            self.api_object.bus3 = None
            self.api_object.V3 = 0
            self.connection_lines[2] = None
            self.api_object.winding3.bus_to = None
            self.terminals[2].setZValue(-1)
        else:
            raise Exception('Unsupported winding index {}'.format(i))

    def arrange_children(self) -> None:
        """
        this function is necessary because this graphic item behaves like a bus,
        but the function itself does nothing
        """
        pass

    def set_tile_color(self, brush: QColor):
        """
        Set the voltage colour
        :param brush: QColor object
        """
        for w in self.winding_circles:
            w.setPen(QPen(brush, self.pen_width, self.style))

    def set_winding_color(self, i, color: QColor):
        """
        Set a winding (loading) colour
        :param i: winding index 0-2
        :param color: QColor
        """
        self.winding_circles[i].setPen(QPen(color, self.pen_width, self.style))
        self.terminals[i].setPen(QPen(color, self.pen_width, self.style))

    def delete(self):
        """
        Remove this element
        @return:
        """
        self.editor.delete_with_dialogue(selected=[self], delete_from_db=False)

    def edit(self):
        """
        Open the appropriate editor dialogue
        :return:
        """
        Sbase = self.editor.circuit.Sbase
        dlg = Transformer3WEditor(self.api_object, Sbase, modify_on_accept=True)
        if dlg.exec():
            pass



# MIT License
#
# Copyright (c) 2021-2022 Yunosuke Ohsugi
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import atexit
import os
import platform

# import darkdetect
import VeraGrid.ThirdParty.qdarktheme as qdarktheme
from VeraGrid.ThirdParty.qdarktheme._style_loader import load_palette, load_stylesheet

_listener = None
_proxy_style = None


def _apply_style(app, additional_qss: str | None, **kargs) -> None:
    from VeraGrid.ThirdParty.qdarktheme._proxy_style import QDarkThemeStyle

    stylesheet = load_stylesheet(**kargs)
    if additional_qss is not None:
        stylesheet += additional_qss
    app.setStyleSheet(stylesheet)

    app.setPalette(
        load_palette(
            kargs["theme"],
            kargs["custom_colors"],
            for_stylesheet=True,
            default_theme=kargs["default_theme"],
        )
    )

    global _proxy_style
    if _proxy_style is None:
        _proxy_style = QDarkThemeStyle()
        app.setStyle(_proxy_style)


def _sync_theme_with_system(app, callback) -> None:
    from VeraGrid.ThirdParty.qdarktheme._os_appearance import listener

    global _listener
    if _listener is not None:
        _listener.sig_run.emit(True)
        return

    _listener = listener.OSThemeSwitchListener(callback)

    if platform.system() == "Darwin":
        app.installEventFilter(_listener)
    else:
        atexit.register(_listener.kill)
        _listener.start()


def enable_hi_dpi() -> None:
    """Allow to HiDPI.

    This function must be set before instantiation of QApplication..
    For Qt6 bindings, HiDPI “just works” without using this function.
    """
    from VeraGrid.ThirdParty.qdarktheme.qtpy.QtCore import Qt
    from VeraGrid.ThirdParty.qdarktheme.qtpy.QtGui import QGuiApplication

    # if hasattr(Qt.ApplicationAttribute, "AA_UseHighDpiPixmaps"):
    #     QGuiApplication.setAttribute(Qt.ApplicationAttribute.AA_UseHighDpiPixmaps)  # type: ignore
    # if hasattr(Qt.ApplicationAttribute, "AA_EnableHighDpiScaling"):
    #     QGuiApplication.setAttribute(Qt.ApplicationAttribute.AA_EnableHighDpiScaling)  # type: ignore
    if hasattr(Qt, "HighDpiScaleFactorRoundingPolicy"):
        os.environ["QT_ENABLE_HIGHDPI_SCALING"] = "1"
        QGuiApplication.setHighDpiScaleFactorRoundingPolicy(
            Qt.HighDpiScaleFactorRoundingPolicy.PassThrough
        )


def stop_sync() -> None:
    """Stop sync with system theme."""
    from VeraGrid.ThirdParty.qdarktheme.qtpy.QtCore import QCoreApplication

    app = QCoreApplication.instance()
    global _listener
    if not app or not _listener:
        return
    _listener.sig_run.emit(False)


def setup_theme(
    theme: str = "dark",
    corner_shape: str = "rounded",
    custom_colors: dict[str, str | dict[str, str]] | None = None,
    additional_qss: str | None = None,
    *,
    default_theme: str = "dark",
) -> None:
    """Apply the theme which looks like flat design to the Qt App completely.

    This function applies the complete style to your Qt application. If the argument theme is ``auto``,
    try to listen to changes to the OS's theme and switch the application theme accordingly.

    Args:
        theme: The theme name. There are `dark`, `light` and `auto`.
            If ``auto``, try to sync with your OS's theme and accent (accent is only on Mac).
            If failed to detect OS's theme, use the default theme set in argument ``default_theme``.
            When primary color(``primary``) or primary child colors
            (such as ``primary>selection.background``) are set to custom_colors,
            disable to sync with the accent.
        corner_shape: The corner shape. There are `rounded` and `sharp` shape.
        custom_colors: The custom color map. Overrides the default color for color id you set.
            Also you can customize a specific theme only. See example 5.
        additional_qss: Additional stylesheet text. You can add your original stylesheet text.
        default_theme: The default theme name.
            The theme set by this argument will be used when system theme detection fails.

    Raises:
        ValueError: If the argument is wrong.
        KeyError: If the color id of custom_colors is wrong.

    Returns:
        The stylesheet string for the given arguments.

    Examples:
        Set stylesheet to your Qt application.

        1. Setup style and sync to system theme ::

            app = QApplication([])
            qdarktheme.setup_theme()

        2. Use Dark Theme ::

            app = QApplication([])
            qdarktheme.setup_theme("dark")

        3. Sharp corner ::

            # Change corner shape to sharp.
            app = QApplication([])
            qdarktheme.setup_theme(corner_shape="sharp")

        4. Customize color ::

            app = QApplication([])
            qdarktheme.setup_theme(custom_colors={"primary": "#D0BCFF"})

        5. Customize a specific theme only ::

            app = QApplication([])
            qdarktheme.setup_theme(
                theme="auto",
                custom_colors={
                    "[dark]": {
                        "primary": "#D0BCFF",
                    }
                },
            )
    """
    from VeraGrid.ThirdParty.qdarktheme.qtpy.QtCore import QCoreApplication

    app = QCoreApplication.instance()
    if not app:
        raise Exception("setup_theme() must be called after instantiation of QApplication.")
    if theme != "auto":
        stop_sync()
    app.setProperty("_qdarktheme_use_setup_style", True)

    def callback():
        _apply_style(
            app,
            additional_qss,
            theme=theme,
            corner_shape=corner_shape,
            custom_colors=custom_colors,
            default_theme=default_theme,
        )

    callback()

    if theme == "auto" and qdarktheme.theme() is not None:
        _sync_theme_with_system(app, callback)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

import re

from PySide6.QtGui import QSyntaxHighlighter, QBrush
from PySide6.QtGui import QTextCharFormat, QColor, QFont


class PythonHighlighter(QSyntaxHighlighter):
    """
    PythonHighlighter
    """

    def __init__(self, document):

        super().__init__(document)

        # Define RGB colors for a dark theme
        cyan = QColor(0, 183, 235)
        magenta = QColor(233, 30, 99)
        # yellow = QColor(255, 235, 59)
        red = QColor(244, 67, 54)
        dark_gray = QColor(96, 125, 139)
        dark_green = QColor(76, 175, 80)
        green = QColor(139, 195, 74)
        light_blue = QColor(33, 150, 243)
        orange = QColor(255, 152, 0)
        purple = QColor(156, 39, 176)
        pink = QColor(233, 30, 99)
        teal = QColor(0, 150, 136)
        deep_purple = QColor(103, 58, 183)
        amber = QColor(255, 193, 7)

        self.highlight_rules = [
            (r'\bdef\b\s+([a-zA-Z_]\w*)', cyan, False),  # Highlight function names
            (r'\bclass\b\s+([a-zA-Z_]\w*)', magenta, False),  # Highlight class names
            (r'\bif\b|\belse\b|\bwhile\b|\bfor\b|\bin\b|\bbreak\b|'
             r'\bcontinue\b|\bpass\b|\btry\b|\bexcept\b|\bfinally\b|\bassert\b', light_blue, True),
            (r'\bimport\b|\bfrom\b|\bas\b|\bglobal\b|\bnonlocal\b|\breturn\b|\byield\b|\bwith\b', teal, True),
            (r'\b|\band\b|\bor\b|\bnot\b|\bin\b', teal, True),
            (r'\bTrue\b|\bFalse', magenta, True),
            (r'\bNone\b|\bint\b|\bfloat\b|\bstr\b|\blist\b|\btuple\b|\bset\b|\bdict\b|\blen\b', deep_purple, True),
            (r'\binput\b|\bopen\b|\bmin\b|\bmax\b|\bsum\b|\babs\b|\bround\b|\bord\b|\bchr\b|\brange\b', amber, True),
            (r'\ball\b|\bany\b|\bzip\b|\benumerate\b|\bsorted\b', pink, True),
            (r'#.*$', dark_gray, False),
            (r'".*?"', pink, False),
            (r"'.*?'", pink, False),
            (r'\b(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?\b', dark_green, False),
            (r'\bprint\b', green, True),  # Highlight 'print' in green
            (r'\+', purple, False),  # Highlight math operations
            (r'-', purple, False),
            (r'\*', purple, False),
            (r'/', purple, False),
            (r'def', red, True),
            (r'=', orange, False),
            (r'<', orange, False),
            (r'>', orange, False),
            (r'class', red, True),
            (r'self', dark_green, False),
            # Add more highlighting rules as needed
        ]

    def highlightBlock(self, text):
        """
        Search the text to highlight
        :param text: some text to highlight
        """
        for pattern, color, bold in self.highlight_rules:
            frmt = QTextCharFormat()
            frmt.setForeground(QBrush(color))

            if bold:
                frmt.setFontWeight(QFont.Weight.Bold)

            for match in re.finditer(pattern, text, re.IGNORECASE):
                start, end = match.span()
                self.setFormat(start, end - start, frmt)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

import sys
from PySide6 import QtWidgets

from VeraGrid.Gui.gui_functions import get_icon_list_model
from VeraGrid.Gui.RmsModelEditor.rms_model_editor import Ui_MainWindow
from VeraGrid.Gui.RmsModelEditor.block_editor import BlockEditor
from VeraGridEngine.Utils.Symbolic.block import Block
import VeraGridEngine.Devices as dev


class RmsModelEditorGUI(QtWidgets.QMainWindow):

    def __init__(self, model: Block, parent=None, ):
        """

        :param parent:
        """
        QtWidgets.QMainWindow.__init__(self, parent)
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)
        self.setWindowTitle('RMS Model editor')

        self.model = model

        self._list_mdl = get_icon_list_model(
            lst=[
                ("Index dynamic parameters", ":/Icons/icons/dyn.svg"),
                ("Numeric dynamic parameters", ":/Icons/icons/dyn.svg"),
                ("External dynamic parameters", ":/Icons/icons/dyn.svg"),
                ("State variables", ":/Icons/icons/dyn.svg"),
                ("Algebraic variables", ":/Icons/icons/dyn.svg"),
                ("External state variables", ":/Icons/icons/dyn.svg"),
                ("External algebraic variables", ":/Icons/icons/dyn.svg")
            ]
        )
        self.ui.listView.setModel(self._list_mdl)

        self.editor = BlockEditor()
        self.ui.editorLayout.addWidget(self.editor)

        self.ui.actionCheckModel.triggered.connect(self.extract_dae)

    def extract_dae(self):
        eqs = self.editor.run()

        for eq in eqs:
            print(str(eq))


if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)
    md = dev.DynamicModel()
    window = RmsModelEditorGUI(md)
    window.resize(1.61 * 700.0, 600.0)  # golden ratio
    window.show()
    sys.exit(app.exec())



# MIT License
#
# Copyright (c) 2021-2022 Yunosuke Ohsugi
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module loading QPalette."""
from __future__ import annotations

from functools import partial

from VeraGrid.ThirdParty.qdarktheme._template.engine import Template


def q_palette(mk_template: partial[Template], color_map: dict[str, str | dict], for_stylesheet: bool):
    """Generate QPalette."""
    from VeraGrid.ThirdParty.qdarktheme.qtpy.QtGui import QColor, QPalette

    def _mk_q_color(text: str):
        template = mk_template(text)
        color_format = template.render(color_map)
        return QColor(color_format)

    palette = QPalette()

    if not for_stylesheet:
        # base
        palette.setColor(QPalette.ColorRole.WindowText, _mk_q_color("{{ foreground|color|palette }}"))
        palette.setColor(
            QPalette.ColorRole.Button, _mk_q_color("{{ treeSectionHeader.background|color|palette }}")
        )
        palette.setColor(QPalette.ColorRole.ButtonText, _mk_q_color("{{ primary|color|palette }}"))
        palette.setColor(QPalette.ColorRole.Base, _mk_q_color("{{ background|color|palette }}"))
        palette.setColor(QPalette.ColorRole.Window, _mk_q_color("{{ background|color|palette }}"))
        palette.setColor(QPalette.ColorRole.Highlight, _mk_q_color("{{ primary|color|palette }}"))
        palette.setColor(
            QPalette.ColorRole.HighlightedText, _mk_q_color("{{ background|color|palette }}")
        )
        palette.setColor(
            QPalette.ColorRole.AlternateBase,
            _mk_q_color("{{ list.alternateBackground|color|palette }}"),
        )
        palette.setColor(
            QPalette.ColorRole.ToolTipBase, _mk_q_color('{{ background|color(state="popup")|palette }}')
        )
        palette.setColor(QPalette.ColorRole.ToolTipText, _mk_q_color("{{ foreground|color|palette }}"))
        if hasattr(QPalette.ColorRole, "Foreground"):
            palette.setColor(
                QPalette.ColorRole.Foreground,  # type: ignore
                _mk_q_color("{{ foreground|color|palette }}"),
            )

        palette.setColor(QPalette.ColorRole.Light, _mk_q_color("{{ border|color|palette }}"))
        palette.setColor(QPalette.ColorRole.Midlight, _mk_q_color("{{ border|color|palette }}"))
        palette.setColor(QPalette.ColorRole.Dark, _mk_q_color("{{ background|color|palette }}"))
        palette.setColor(QPalette.ColorRole.Mid, _mk_q_color("{{ border|color|palette }}"))
        palette.setColor(QPalette.ColorRole.Shadow, _mk_q_color("{{ border|color|palette }}"))

        # disabled
        palette.setColor(
            QPalette.ColorGroup.Disabled,
            QPalette.ColorRole.WindowText,
            _mk_q_color('{{ foreground|color(state="disabled")|palette }}'),
        )
        palette.setColor(
            QPalette.ColorGroup.Disabled,
            QPalette.ColorRole.ButtonText,
            _mk_q_color('{{ foreground|color(state="disabled")|palette }}'),
        )
        palette.setColor(
            QPalette.ColorGroup.Disabled,
            QPalette.ColorRole.Highlight,
            _mk_q_color('{{ foreground|color(state="disabledSelectionBackground")|palette }}'),
        )
        palette.setColor(
            QPalette.ColorGroup.Disabled,
            QPalette.ColorRole.HighlightedText,
            _mk_q_color('{{ foreground|color(state="disabled")|palette }}'),
        )

        # inactive
        palette.setColor(
            QPalette.ColorGroup.Inactive,
            QPalette.ColorRole.Highlight,
            _mk_q_color('{{ primary|color(state="list.inactiveSelectionBackground")|palette }}'),
        )
        palette.setColor(
            QPalette.ColorGroup.Inactive,
            QPalette.ColorRole.HighlightedText,
            _mk_q_color("{{ foreground|color|palette }}"),
        )

    palette.setColor(
        QPalette.ColorRole.Text, _mk_q_color('{{ foreground|color(state="icon")|palette }}')
    )
    palette.setColor(
        QPalette.ColorGroup.Disabled,
        QPalette.ColorRole.Text,
        _mk_q_color('{{ foreground|color(state="disabled")|palette }}'),
    )
    palette.setColor(QPalette.ColorRole.Link, _mk_q_color("{{ primary|color|palette }}"))
    palette.setColor(QPalette.ColorRole.LinkVisited, _mk_q_color("{{ linkVisited|color|palette }}"))
    if hasattr(QPalette.ColorRole, "PlaceholderText"):
        palette.setColor(
            QPalette.ColorRole.PlaceholderText,
            _mk_q_color('{{ foreground|color(state="input.placeholder")|palette }}'),
        )

    palette.setColor(
        QPalette.ColorGroup.Disabled,
        QPalette.ColorRole.Link,
        _mk_q_color('{{ foreground|color(state="disabledSelectionBackground")|palette }}'),
    )
    palette.setColor(
        QPalette.ColorGroup.Disabled,
        QPalette.ColorRole.LinkVisited,
        _mk_q_color('{{ foreground|color(state="disabled")|palette }}'),
    )

    return palette



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
from __future__ import annotations
from typing import Union, TYPE_CHECKING, Dict, Callable, List
from PySide6.QtCore import QPointF
from PySide6.QtGui import QBrush

from VeraGrid.Gui.Diagrams.MapWidget.Branches.map_line_segment import MapLineSegment
from VeraGrid.Gui.Diagrams.generic_graphics import GenericDiagramWidget

from VeraGridEngine.Devices.types import ALL_DEV_TYPES

if TYPE_CHECKING:  # Only imports the below statements during type checking
    from VeraGrid.Gui.Diagrams.MapWidget.grid_map_widget import GridMapWidget


class NodeTemplate(GenericDiagramWidget):
    """
    Node Template
    """

    def __init__(self,
                 api_object: ALL_DEV_TYPES = None,
                 editor: Union[GridMapWidget, None] = None,
                 draw_labels: bool = True,
                 lat: float = 0.0,
                 lon: float = 0.0):
        """

        :param api_object: any VeraGrid device object
        :param editor: GridMapWidget
        :param draw_labels:
        :param lat:
        :param lon:
        """
        GenericDiagramWidget.__init__(self,
                                      parent=None,
                                      api_object=api_object,
                                      editor=editor,
                                      draw_labels=draw_labels)

        self.lat = lat
        self.lon = lon

        self.hovered = False
        self.needsUpdate = False

        self._hosting_connections: Dict[MapLineSegment, Callable[[float, float], None]] = dict()

    def add_position_change_callback(self, obj: MapLineSegment, fcn: Callable[[float, float], None]):
        """
        Add callable function
        :param fcn: Function that accepts two floats (x, y) to update the positions
        :return:
        """
        self._hosting_connections[obj] = fcn

    def set_callbacks(self, x: float, y: float) -> None:
        """
        Call all callback functions with x, y
        :param x: x position
        :param y: y position
        """
        for obj, fcn in self._hosting_connections.items():
            fcn(x, y)

    def valid_coordinates(self) -> bool:
        """
        Checks if the coordinates are different from 0, 0
        :return: ok?
        """
        return self.lon != 0.0 and self.lat != 0.0

    def getRealPos(self):
        """

        :return:
        """
        return QPointF(0, 0)

    def setBrush(self, brush: QBrush):
        """
        Dummy placeholder function
        :param brush:
        :return:
        """
        pass

    def delete_hosting_connection(self, graphic_obj: MapLineSegment):
        """
        Delete object graphically connected to the graphical bus
        :param graphic_obj: LineGraphicTemplateItem (or child of this)
        """
        if graphic_obj in self._hosting_connections.keys():
            del self._hosting_connections[graphic_obj]
        else:
            print(f'No such hosting connection for {graphic_obj}')

    def get_hosting_line_segments(self) -> List[MapLineSegment]:
        """
        Ger a list of the lines that are hosted here
        """
        return list(self._hosting_connections.keys())


# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
import numpy as np
import pandas as pd
from typing import List
from PySide6 import QtWidgets
from matplotlib import pyplot as plt

import VeraGrid.Gui.gui_functions as gf
from VeraGridEngine.basic_structures import Logger
from VeraGridEngine.enumerations import DeviceType
from VeraGridEngine.Devices.types import ALL_DEV_TYPES
from VeraGrid.Gui.general_dialogues import NewProfilesStructureDialogue, TimeReIndexDialogue, LogsDialogue
from VeraGrid.Gui.messages import yes_no_question, warning_msg, info_msg
from VeraGrid.Gui.Main.SubClasses.Model.data_base import DataBaseTableMain
from VeraGrid.Gui.ProfilesInput.models_dialogue import ModelsInputGUI
from VeraGrid.Gui.ProfilesInput.profile_dialogue import ProfileInputGUI, GeneratorsProfileOptionsDialogue
from VeraGrid.Gui.profiles_model import ProfilesModel


class TimeEventsMain(DataBaseTableMain):
    """
    Diagrams Main
    """

    def __init__(self, parent=None):
        """

        @param parent:
        """

        # create main window
        DataBaseTableMain.__init__(self, parent)

        # --------------------------------------------------------------------------------------------------------------
        self.ui.actionre_index_time.triggered.connect(self.re_index_time)

        # Buttons
        self.ui.new_profiles_structure_pushButton.clicked.connect(self.new_profiles_structure)
        self.ui.delete_profiles_structure_pushButton.clicked.connect(self.delete_profiles_structure)
        self.ui.edit_profiles_pushButton.clicked.connect(self.import_profiles)
        self.ui.edit_profiles_from_models_pushButton.clicked.connect(self.import_profiles_from_models)
        self.ui.set_profile_state_button.clicked.connect(self.set_profile_state_to_snapshot)
        self.ui.profile_add_pushButton.clicked.connect(lambda: self.modify_profiles('+'))
        self.ui.profile_subtract_pushButton.clicked.connect(lambda: self.modify_profiles('-'))
        self.ui.profile_multiply_pushButton.clicked.connect(lambda: self.modify_profiles('*'))
        self.ui.profile_divide_pushButton.clicked.connect(lambda: self.modify_profiles('/'))
        self.ui.set_profile_value_pushButton.clicked.connect(lambda: self.modify_profiles('set'))
        self.ui.set_linear_combination_profile_pushButton.clicked.connect(self.set_profile_as_linear_combination)
        self.ui.plot_time_series_pushButton.clicked.connect(self.plot_profiles)
        self.ui.copy_profile_pushButton.clicked.connect(self.copy_profiles)
        self.ui.paste_profiles_pushButton.clicked.connect(self.paste_profiles)

        # combobox change
        self.ui.device_type_magnitude_comboBox.currentTextChanged.connect(self.profile_device_type_changed)

    def profile_device_type_changed(self):
        """
        profile_device_type_changed
        """
        self.display_profiles(proxy_mdl=self.get_current_objects_model_view())

    def new_profiles_structure(self):
        """
        Create new profiles structure
        :return:
        """
        dlg = NewProfilesStructureDialogue()
        if dlg.exec():
            steps, step_length, step_unit, time_base = dlg.get_values()

            self.ui.profiles_tableView.setModel(None)

            self.circuit.create_profiles(steps=steps,
                                         step_length=step_length,
                                         step_unit=step_unit,
                                         time_base=time_base)

            self.display_profiles(proxy_mdl=self.get_current_objects_model_view())

            self.update_date_dependent_combos()

    def delete_profiles_structure(self):
        """
        Delete all profiles
        :return: Nothing
        """

        if self.circuit.time_profile is not None:
            quit_msg = "Are you sure that you want to delete the profiles?"
            reply = QtWidgets.QMessageBox.question(self, 'Message', quit_msg,
                                                   QtWidgets.QMessageBox.StandardButton.Yes,
                                                   QtWidgets.QMessageBox.StandardButton.No)

            if reply == QtWidgets.QMessageBox.StandardButton.Yes.value:
                self.circuit.delete_profiles()
                self.ui.profiles_tableView.setModel(None)
                self.update_date_dependent_combos()
                self.update_from_to_list_views()
            else:
                pass
        else:
            warning_msg('There are no profiles', 'Delete profiles')

    def import_profiles(self):
        """
        Profile importer
        """
        dev_type_text = self.get_db_object_selected_type()

        if dev_type_text is not None:
            idx = self.ui.device_type_magnitude_comboBox.currentIndex()

            magnitudes, mag_types = self.circuit.profile_magnitudes[dev_type_text]
            dev_type = self.circuit.device_type_name_dict[dev_type_text]
            objects: List[ALL_DEV_TYPES] = self.circuit.get_elements_by_type(dev_type)
            magnitude = magnitudes[idx]

            if len(objects) > 0 and idx > -1:
                self.profile_input_dialogue = ProfileInputGUI(parent=self,
                                                              circuit=self.circuit,
                                                              dev_type=dev_type,
                                                              objects=objects,
                                                              magnitude=magnitude)

                self.profile_input_dialogue.resize(int(1.61 * 600.0), 550)  # golden ratio
                self.profile_input_dialogue.exec()  # exec leaves the parent on hold

                # Note: the ProfileInputGUI will handle the profile assigning

                if self.profile_input_dialogue.was_accepted:

                    # set up sliders
                    self.update_date_dependent_combos()
                    self.display_profiles(proxy_mdl=self.get_current_objects_model_view())
                    self.show_info_toast("Profiles imported", duration=3000)

                    # ask to update active profile when magnitude is P for generators and loads
                    if len(objects) > 0:
                        if magnitude == 'P':
                            if objects[0].device_type == DeviceType.GeneratorDevice:

                                dlg = GeneratorsProfileOptionsDialogue()
                                dlg.exec()

                                if dlg.correct_active_profile.isChecked():
                                    self.fix_generators_active_based_on_the_power(ask_before=False)
                                    self.show_info_toast("Generators active status set")

                                if dlg.set_non_dispatchable.isChecked():
                                    for i, elm in enumerate(objects):
                                        if self.profile_input_dialogue.has_profile(i):
                                            # if there was a profile, we want the generator not dispatchable
                                            elm.enabled_dispatch = False
                                        else:
                                            elm.enabled_dispatch = True

                                    self.show_info_toast("Generators dispatchable status set")

                            elif objects[0].device_type == DeviceType.LoadDevice:
                                ok1 = yes_no_question("Do you want to correct the loads active profile "
                                                      "based on the active power profile?",
                                                      "Match")
                                if ok1:
                                    self.fix_loads_active_based_on_the_power(ask_before=False)
                                    self.show_info_toast("Loads active status set")

                    else:
                        # the dialogue was closed
                        self.show_warning_toast("No profiles imported...")
                else:
                    # the dialogue was closed
                    self.show_warning_toast("No profiles imported...")

            else:
                self.show_error_toast("There are no objects...", duration=3000)

    def modify_profiles(self, operation='+'):
        """
        Edit profiles with a linear combination
        Args:
            operation: '+', '-', '*', '/'

        Returns: Nothing
        """
        value = self.ui.profile_factor_doubleSpinBox.value()

        dev_type_text = self.get_db_object_selected_type()

        if dev_type_text is not None:
            magnitudes, mag_types = self.circuit.profile_magnitudes[dev_type_text]
            idx = self.ui.device_type_magnitude_comboBox.currentIndex()
            magnitude = magnitudes[idx]

            dev_type = self.circuit.device_type_name_dict[dev_type_text]
            objects: List[ALL_DEV_TYPES] = self.circuit.get_elements_by_type(dev_type)
            # Assign profiles
            if len(objects) > 0:

                indices = self.ui.profiles_tableView.selectedIndexes()

                # attr = objects[0].properties_with_profile[magnitude]

                model = self.ui.profiles_tableView.model()

                mod_cols = list()

                if len(indices) == 0:
                    # no index was selected
                    for i, elm in enumerate(objects):

                        # get the property object
                        gc_prop = elm.registered_properties[magnitude]

                        # get the profile
                        profile = elm.get_profile_by_prop(prop=gc_prop)

                        # compute the dense array (this is the simple way of doing this)
                        array = profile.toarray()

                        if operation == '+':
                            mod_array = (array + value).astype(gc_prop.tpe)
                            mod_cols.append(i)

                        elif operation == '-':
                            mod_array = (array - value).astype(gc_prop.tpe)
                            mod_cols.append(i)

                        elif operation == '*':
                            mod_array = (array * value).astype(gc_prop.tpe)
                            mod_cols.append(i)

                        elif operation == '/':
                            mod_array = (array / value).astype(gc_prop.tpe)
                            mod_cols.append(i)

                        elif operation == 'set':
                            mod_array = (np.ones(len(array)) * value).astype(gc_prop.tpe)
                            mod_cols.append(i)

                        else:
                            raise Exception('Operation not supported: ' + str(operation))

                        # apply the newly computed array
                        profile.set(arr=mod_array)

                else:
                    # indices were selected ...

                    for idx in indices:

                        # get the device
                        elm = objects[idx.column()]

                        # get the property object
                        gc_prop = elm.registered_properties[magnitude]

                        # get the profile
                        profile = elm.get_profile_by_prop(prop=gc_prop)

                        # compute the dense array (this is the simple way of doing this)
                        array = profile.toarray().copy()

                        if operation == '+':
                            array[idx.row()] += value
                            mod_cols.append(idx.column())

                        elif operation == '-':
                            array[idx.row()] -= value
                            mod_cols.append(idx.column())

                        elif operation == '*':
                            array[idx.row()] *= value
                            mod_cols.append(idx.column())

                        elif operation == '/':
                            array[idx.row()] /= value
                            mod_cols.append(idx.column())

                        elif operation == 'set':
                            array[idx.row()] = value
                            mod_cols.append(idx.column())

                        else:
                            raise Exception('Operation not supported: ' + str(operation))

                        # apply the newly computed array
                        profile.set(arr=array)

                # update model
                model.update()

    def set_profile_as_linear_combination(self):
        """
        Edit profiles with a linear combination
        Returns: Nothing
        """
        logger: Logger = Logger()
        # value = self.ui.profile_factor_doubleSpinBox.value()

        dev_type_text = self.get_db_object_selected_type()
        if dev_type_text is not None:
            magnitudes, mag_types = self.circuit.profile_magnitudes[dev_type_text]
            idx_from = self.ui.device_type_magnitude_comboBox.currentIndex()
            magnitude_from = magnitudes[idx_from]

            idx_to = self.ui.device_type_magnitude_comboBox_2.currentIndex()
            magnitude_to = magnitudes[idx_to]

            if self.circuit.valid_for_simulation() and magnitude_from != magnitude_to:

                msg = "Are you sure that you want to overwrite the values " + magnitude_to + \
                      " with the values of " + magnitude_from + "?"

                reply = QtWidgets.QMessageBox.question(self, 'Message', msg,
                                                       QtWidgets.QMessageBox.StandardButton.Yes,
                                                       QtWidgets.QMessageBox.StandardButton.No)

                if reply == QtWidgets.QMessageBox.StandardButton.Yes.value:

                    dev_type = self.circuit.device_type_name_dict[dev_type_text]
                    objects: List[ALL_DEV_TYPES] = self.circuit.get_elements_by_type(dev_type)

                    # Assign profiles
                    if len(objects) > 0:

                        for i, elm in enumerate(objects):
                            profile_from = elm.get_profile(magnitude=magnitude_from)
                            profile_to = elm.get_profile(magnitude=magnitude_to)
                            if profile_from is not None and profile_to is not None:
                                profile_to.set(profile_from.toarray())
                            else:
                                print(f"P or Q profile None in {elm.name}")

                        self.display_profiles(proxy_mdl=self.get_current_objects_model_view())

                else:
                    # rejected the operation
                    pass

            else:
                # no buses or no actual change
                pass

        if logger.has_logs():
            dlg = LogsDialogue("Set profile", logger=logger)
            dlg.exec()

    def re_index_time(self):
        """
        Re-index time
        :return:
        """

        dlg = TimeReIndexDialogue()
        dlg.setModal(True)
        dlg.exec()

        if dlg.is_accepted:
            self.circuit.re_index_time2(t0=dlg.date_time_editor.dateTime().toPython(),
                                        step_size=dlg.step_length.value(),
                                        step_unit=dlg.units.currentText())

            self.update_date_dependent_combos()

    def plot_profiles(self):
        """
        Plot profiles from the time events
        """
        dev_type_text = self.get_db_object_selected_type()

        if dev_type_text is not None:
            magnitudes, mag_types = self.circuit.profile_magnitudes[dev_type_text]
            idx = self.ui.device_type_magnitude_comboBox.currentIndex()
            magnitude = magnitudes[idx]

            dev_type = self.circuit.device_type_name_dict[dev_type_text]
            objects = self.circuit.get_elements_by_type(dev_type)

            # get the selected element
            obj_idx = self.ui.profiles_tableView.selectedIndexes()

            t = self.circuit.time_profile

            # Assign profiles
            if len(obj_idx):
                fig = plt.figure(figsize=(12, 8))
                ax = fig.add_subplot(111)

                k = obj_idx[0].column()
                units_dict = {attr: pair.units for attr, pair in objects[k].registered_properties.items()}

                unit = units_dict[magnitude]
                ax.set_ylabel(unit)

                # get the unique columns in the selected cells
                cols = set()
                for i in range(len(obj_idx)):
                    cols.add(obj_idx[i].column())

                # plot every column
                dta = dict()
                for k in cols:
                    dta[objects[k].name] = objects[k].get_profile(magnitude=magnitude).toarray()
                df = pd.DataFrame(data=dta, index=t)
                df.plot(ax=ax)

                plt.show()

    def import_profiles_from_models(self):
        """
        Open the dialogue to load profile data from models
        """

        if not self.circuit.valid_for_simulation():
            warning_msg("There are no objects to which to assign a profile. \n"
                        "You need to load or create a grid!")
            return

        if self.circuit.time_profile is None:
            self.models_input_dialogue = ModelsInputGUI(parent=self)

            self.models_input_dialogue.resize(int(1.61 * 600.0), 550)  # golden ratio
            self.models_input_dialogue.exec()  # exec leaves the parent on hold

            if self.models_input_dialogue.grids_model is not None:

                logger = Logger()
                self.models_input_dialogue.process(main_grid=self.circuit, logger=logger)

                # set up sliders
                self.update_date_dependent_combos()
                self.display_profiles(proxy_mdl=self.get_current_objects_model_view())

                if logger.has_logs():
                    dialogue = LogsDialogue(name="Import profiles", logger=logger)
                    dialogue.exec()

        else:
            warning_msg("The import of profiles from many grid models "
                        "can only be done if the grid has not profiles :/")

    def get_circuit_snapshot_datetime(self):
        """
        Set the datetime from the circuit
        """
        val = self.circuit.snapshot_time
        self.ui.snapshot_dateTimeEdit.setDateTime(val)

    def set_profile_state_to_snapshot(self):
        """
        Set the selected profiles state in the grid
        """
        idx = self.ui.db_step_slider.value()

        if idx > -1:
            self.circuit.set_state(t=idx)
            self.get_circuit_snapshot_datetime()
            self.show_info_toast("Profile value set to the snapshot")
        else:
            info_msg('Select a time series step to copy to the snapshot', 'Set snapshot')

    def copy_profiles(self):
        """
        Copy the current displayed profiles to the clipboard
        """

        mdl: ProfilesModel = self.ui.profiles_tableView.model()

        cols = set()
        if len(self.ui.profiles_tableView.selectedIndexes()) > 0:
            for index in self.ui.profiles_tableView.selectedIndexes():
                row_idx = index.row()
                col_idx = index.column()
                cols.add(col_idx)
        else:
            row_idx = 0
            col_idx = 0

        if mdl is not None:
            mdl.copy_to_clipboard(cols=list(cols))
        else:
            warning_msg('There is no profile displayed, please display one', 'Copy profile to clipboard')

    def paste_profiles(self):
        """
        Paste clipboard data into the profile
        """

        mdl = self.ui.profiles_tableView.model()
        if mdl is not None:

            if len(self.ui.profiles_tableView.selectedIndexes()) > 0:
                index = self.ui.profiles_tableView.selectedIndexes()[0]
                row_idx = index.row()
                col_idx = index.column()
            else:
                row_idx = 0
                col_idx = 0

            mdl.paste_from_clipboard(row_idx=row_idx, col_idx=col_idx)
        else:
            warning_msg('There is no profile displayed, please display one', 'Paste profile to clipboard')



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.  
# SPDX-License-Identifier: MPL-2.0


import os
import pandas as pd
from VeraGridEngine.Devices.Branches.line import SequenceLineType, UndergroundLineType
from VeraGridEngine.Devices.Branches.transformer import TransformerType
from VeraGridEngine.Devices.Branches.wire import Wire
from VeraGridEngine.IO.veragrid.catalogue import parse_transformer_types, parse_cable_types, parse_wire_types, parse_sequence_line_types


def get_transformer_catalogue():
    """

    :return:
    """
    here = os.path.dirname(os.path.abspath(__file__))
    fname = os.path.join(here, 'data', 'transformers.csv')

    if os.path.exists(fname):
        df = pd.read_csv(fname)

        return parse_transformer_types(df)
    else:
        return list()


def get_cables_catalogue():
    """

    :return:
    """
    here = os.path.dirname(os.path.abspath(__file__))
    fname = os.path.join(here, 'data', 'cables.csv')

    if os.path.exists(fname):
        df = pd.read_csv(fname)

        return parse_cable_types(df)
    else:
        return list()


def get_wires_catalogue():
    """

    :return:
    """
    here = os.path.dirname(os.path.abspath(__file__))
    fname = os.path.join(here, 'data', 'wires.csv')

    if os.path.exists(fname):
        df = pd.read_csv(fname)

        return parse_wire_types(df)
    else:
        return list()


def get_sequence_lines_catalogue():
    """

    :return:
    """
    here = os.path.dirname(os.path.abspath(__file__))
    fname = os.path.join(here, 'data', 'sequence_lines.csv')

    if os.path.exists(fname):
        df = pd.read_csv(fname)

        return parse_sequence_line_types(df)
    else:
        return list()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'gui.ui'
##
## Created by: Qt User Interface Compiler version 6.7.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QAction, QBrush, QColor, QConicalGradient,
    QCursor, QFont, QFontDatabase, QGradient,
    QIcon, QImage, QKeySequence, QLinearGradient,
    QPainter, QPalette, QPixmap, QRadialGradient,
    QTransform)
from PySide6.QtWidgets import (QApplication, QCheckBox, QDoubleSpinBox, QFrame,
    QGridLayout, QHBoxLayout, QHeaderView, QLabel,
    QMainWindow, QSizePolicy, QSpacerItem, QStatusBar,
    QTabWidget, QToolBar, QTreeView, QVBoxLayout,
    QWidget)
from .icons_rc import *

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        if not MainWindow.objectName():
            MainWindow.setObjectName(u"MainWindow")
        MainWindow.resize(912, 540)
        icon = QIcon()
        icon.addFile(u":/icons/icons/inputs_analysis 2.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        MainWindow.setWindowIcon(icon)
        self.actionSave_diagnostic = QAction(MainWindow)
        self.actionSave_diagnostic.setObjectName(u"actionSave_diagnostic")
        icon1 = QIcon()
        icon1.addFile(u":/icons/icons/savec.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.actionSave_diagnostic.setIcon(icon1)
        self.actionFix_issues = QAction(MainWindow)
        self.actionFix_issues.setObjectName(u"actionFix_issues")
        icon2 = QIcon()
        icon2.addFile(u":/icons/icons/fix.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.actionFix_issues.setIcon(icon2)
        self.actionAnalyze = QAction(MainWindow)
        self.actionAnalyze.setObjectName(u"actionAnalyze")
        self.actionAnalyze.setIcon(icon)
        self.centralwidget = QWidget(MainWindow)
        self.centralwidget.setObjectName(u"centralwidget")
        self.verticalLayout_2 = QVBoxLayout(self.centralwidget)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.tabWidget = QTabWidget(self.centralwidget)
        self.tabWidget.setObjectName(u"tabWidget")
        self.tab = QWidget()
        self.tab.setObjectName(u"tab")
        self.verticalLayout = QVBoxLayout(self.tab)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.logsTreeView = QTreeView(self.tab)
        self.logsTreeView.setObjectName(u"logsTreeView")
        self.logsTreeView.setFrameShape(QFrame.NoFrame)

        self.verticalLayout.addWidget(self.logsTreeView)

        self.tabWidget.addTab(self.tab, icon, "")
        self.tab_2 = QWidget()
        self.tab_2.setObjectName(u"tab_2")
        self.horizontalLayout = QHBoxLayout(self.tab_2)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.frame = QFrame(self.tab_2)
        self.frame.setObjectName(u"frame")
        self.frame.setMaximumSize(QSize(400, 16777215))
        self.frame.setFrameShape(QFrame.NoFrame)
        self.frame.setFrameShadow(QFrame.Raised)
        self.gridLayout = QGridLayout(self.frame)
        self.gridLayout.setObjectName(u"gridLayout")
        self.label_4 = QLabel(self.frame)
        self.label_4.setObjectName(u"label_4")
        palette = QPalette()
        brush = QBrush(QColor(119, 118, 123, 255))
        brush.setStyle(Qt.SolidPattern)
        palette.setBrush(QPalette.Active, QPalette.WindowText, brush)
        palette.setBrush(QPalette.Active, QPalette.Text, brush)
        palette.setBrush(QPalette.Active, QPalette.ButtonText, brush)
        brush1 = QBrush(QColor(0, 0, 0, 255))
        brush1.setStyle(Qt.SolidPattern)
        palette.setBrush(QPalette.Inactive, QPalette.WindowText, brush1)
        palette.setBrush(QPalette.Inactive, QPalette.Text, brush1)
        palette.setBrush(QPalette.Inactive, QPalette.ButtonText, brush1)
        brush2 = QBrush(QColor(190, 190, 190, 255))
        brush2.setStyle(Qt.SolidPattern)
        palette.setBrush(QPalette.Disabled, QPalette.WindowText, brush2)
        palette.setBrush(QPalette.Disabled, QPalette.Text, brush2)
        palette.setBrush(QPalette.Disabled, QPalette.ButtonText, brush2)
        self.label_4.setPalette(palette)
        font = QFont()
        font.setPointSize(14)
        font.setBold(True)
        self.label_4.setFont(font)

        self.gridLayout.addWidget(self.label_4, 3, 0, 1, 1)

        self.label_5 = QLabel(self.frame)
        self.label_5.setObjectName(u"label_5")

        self.gridLayout.addWidget(self.label_5, 11, 0, 1, 1)

        self.transformerVccMaxSpinBox = QDoubleSpinBox(self.frame)
        self.transformerVccMaxSpinBox.setObjectName(u"transformerVccMaxSpinBox")
        self.transformerVccMaxSpinBox.setDecimals(1)
        self.transformerVccMaxSpinBox.setValue(20.000000000000000)

        self.gridLayout.addWidget(self.transformerVccMaxSpinBox, 11, 2, 1, 1)

        self.label = QLabel(self.frame)
        self.label.setObjectName(u"label")

        self.gridLayout.addWidget(self.label, 1, 0, 1, 1)

        self.virtualTapToleranceSpinBox = QDoubleSpinBox(self.frame)
        self.virtualTapToleranceSpinBox.setObjectName(u"virtualTapToleranceSpinBox")
        self.virtualTapToleranceSpinBox.setDecimals(1)
        self.virtualTapToleranceSpinBox.setMaximum(9999.000000000000000)
        self.virtualTapToleranceSpinBox.setValue(10.000000000000000)

        self.gridLayout.addWidget(self.virtualTapToleranceSpinBox, 9, 1, 1, 2)

        self.label_3 = QLabel(self.frame)
        self.label_3.setObjectName(u"label_3")

        self.gridLayout.addWidget(self.label_3, 7, 0, 1, 1)

        self.label_12 = QLabel(self.frame)
        self.label_12.setObjectName(u"label_12")
        palette1 = QPalette()
        palette1.setBrush(QPalette.Active, QPalette.WindowText, brush)
        palette1.setBrush(QPalette.Active, QPalette.Text, brush)
        palette1.setBrush(QPalette.Active, QPalette.ButtonText, brush)
        palette1.setBrush(QPalette.Inactive, QPalette.WindowText, brush1)
        palette1.setBrush(QPalette.Inactive, QPalette.Text, brush1)
        palette1.setBrush(QPalette.Inactive, QPalette.ButtonText, brush1)
        palette1.setBrush(QPalette.Disabled, QPalette.WindowText, brush2)
        palette1.setBrush(QPalette.Disabled, QPalette.Text, brush2)
        palette1.setBrush(QPalette.Disabled, QPalette.ButtonText, brush2)
        self.label_12.setPalette(palette1)
        self.label_12.setFont(font)

        self.gridLayout.addWidget(self.label_12, 0, 0, 1, 1)

        self.label_11 = QLabel(self.frame)
        self.label_11.setObjectName(u"label_11")

        self.gridLayout.addWidget(self.label_11, 2, 0, 1, 1)

        self.verticalSpacer = QSpacerItem(20, 94, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.gridLayout.addItem(self.verticalSpacer, 15, 1, 1, 2)

        self.transformerTapAngleMaxSpinBox = QDoubleSpinBox(self.frame)
        self.transformerTapAngleMaxSpinBox.setObjectName(u"transformerTapAngleMaxSpinBox")
        self.transformerTapAngleMaxSpinBox.setMinimum(-99.000000000000000)
        self.transformerTapAngleMaxSpinBox.setValue(3.140000000000000)

        self.gridLayout.addWidget(self.transformerTapAngleMaxSpinBox, 7, 2, 1, 1)

        self.transformerTapAngleMinSpinBox = QDoubleSpinBox(self.frame)
        self.transformerTapAngleMinSpinBox.setObjectName(u"transformerTapAngleMinSpinBox")
        self.transformerTapAngleMinSpinBox.setMinimum(-99999.000000000000000)
        self.transformerTapAngleMinSpinBox.setMaximum(99999.000000000000000)
        self.transformerTapAngleMinSpinBox.setValue(-3.140000000000000)

        self.gridLayout.addWidget(self.transformerTapAngleMinSpinBox, 7, 1, 1, 1)

        self.label_2 = QLabel(self.frame)
        self.label_2.setObjectName(u"label_2")

        self.gridLayout.addWidget(self.label_2, 5, 0, 1, 1)

        self.label_15 = QLabel(self.frame)
        self.label_15.setObjectName(u"label_15")

        self.gridLayout.addWidget(self.label_15, 14, 0, 1, 1)

        self.transformerTapModuleMinSpinBox = QDoubleSpinBox(self.frame)
        self.transformerTapModuleMinSpinBox.setObjectName(u"transformerTapModuleMinSpinBox")
        self.transformerTapModuleMinSpinBox.setValue(0.950000000000000)

        self.gridLayout.addWidget(self.transformerTapModuleMinSpinBox, 5, 1, 1, 1)

        self.lineNominalVoltageToleranceSpinBox = QDoubleSpinBox(self.frame)
        self.lineNominalVoltageToleranceSpinBox.setObjectName(u"lineNominalVoltageToleranceSpinBox")
        self.lineNominalVoltageToleranceSpinBox.setDecimals(1)
        self.lineNominalVoltageToleranceSpinBox.setMaximum(9999.000000000000000)
        self.lineNominalVoltageToleranceSpinBox.setValue(10.000000000000000)

        self.gridLayout.addWidget(self.lineNominalVoltageToleranceSpinBox, 14, 1, 1, 2)

        self.label_13 = QLabel(self.frame)
        self.label_13.setObjectName(u"label_13")

        self.gridLayout.addWidget(self.label_13, 12, 1, 1, 1)

        self.label_14 = QLabel(self.frame)
        self.label_14.setObjectName(u"label_14")
        palette2 = QPalette()
        palette2.setBrush(QPalette.Active, QPalette.WindowText, brush)
        palette2.setBrush(QPalette.Active, QPalette.Text, brush)
        palette2.setBrush(QPalette.Active, QPalette.ButtonText, brush)
        palette2.setBrush(QPalette.Inactive, QPalette.WindowText, brush1)
        palette2.setBrush(QPalette.Inactive, QPalette.Text, brush1)
        palette2.setBrush(QPalette.Inactive, QPalette.ButtonText, brush1)
        palette2.setBrush(QPalette.Disabled, QPalette.WindowText, brush2)
        palette2.setBrush(QPalette.Disabled, QPalette.Text, brush2)
        palette2.setBrush(QPalette.Disabled, QPalette.ButtonText, brush2)
        self.label_14.setPalette(palette2)
        self.label_14.setFont(font)

        self.gridLayout.addWidget(self.label_14, 13, 0, 1, 1)

        self.label_6 = QLabel(self.frame)
        self.label_6.setObjectName(u"label_6")

        self.gridLayout.addWidget(self.label_6, 9, 0, 1, 1)

        self.genVsetMaxSpinBox = QDoubleSpinBox(self.frame)
        self.genVsetMaxSpinBox.setObjectName(u"genVsetMaxSpinBox")
        self.genVsetMaxSpinBox.setValue(1.050000000000000)

        self.gridLayout.addWidget(self.genVsetMaxSpinBox, 1, 2, 1, 1)

        self.transformerVccMinSpinBox = QDoubleSpinBox(self.frame)
        self.transformerVccMinSpinBox.setObjectName(u"transformerVccMinSpinBox")
        self.transformerVccMinSpinBox.setDecimals(1)
        self.transformerVccMinSpinBox.setMinimum(0.000000000000000)
        self.transformerVccMinSpinBox.setMaximum(100.000000000000000)
        self.transformerVccMinSpinBox.setValue(1.000000000000000)

        self.gridLayout.addWidget(self.transformerVccMinSpinBox, 11, 1, 1, 1)

        self.genVsetMinSpinBox = QDoubleSpinBox(self.frame)
        self.genVsetMinSpinBox.setObjectName(u"genVsetMinSpinBox")
        self.genVsetMinSpinBox.setValue(0.950000000000000)

        self.gridLayout.addWidget(self.genVsetMinSpinBox, 1, 1, 1, 1)

        self.transformerTapModuleMaxSpinBox = QDoubleSpinBox(self.frame)
        self.transformerTapModuleMaxSpinBox.setObjectName(u"transformerTapModuleMaxSpinBox")
        self.transformerTapModuleMaxSpinBox.setValue(1.050000000000000)

        self.gridLayout.addWidget(self.transformerTapModuleMaxSpinBox, 5, 2, 1, 1)

        self.fixTimeSeriesCheckBox = QCheckBox(self.frame)
        self.fixTimeSeriesCheckBox.setObjectName(u"fixTimeSeriesCheckBox")
        self.fixTimeSeriesCheckBox.setChecked(True)

        self.gridLayout.addWidget(self.fixTimeSeriesCheckBox, 16, 0, 1, 3)


        self.horizontalLayout.addWidget(self.frame)

        self.frame_2 = QFrame(self.tab_2)
        self.frame_2.setObjectName(u"frame_2")
        self.frame_2.setFrameShape(QFrame.NoFrame)
        self.frame_2.setFrameShadow(QFrame.Raised)
        self.gridLayout_2 = QGridLayout(self.frame_2)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.label_8 = QLabel(self.frame_2)
        self.label_8.setObjectName(u"label_8")

        self.gridLayout_2.addWidget(self.label_8, 1, 0, 1, 1)

        self.verticalSpacer_2 = QSpacerItem(20, 373, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.gridLayout_2.addItem(self.verticalSpacer_2, 3, 0, 1, 1)

        self.label_16 = QLabel(self.frame_2)
        self.label_16.setObjectName(u"label_16")
        palette3 = QPalette()
        palette3.setBrush(QPalette.Active, QPalette.WindowText, brush)
        palette3.setBrush(QPalette.Active, QPalette.Text, brush)
        palette3.setBrush(QPalette.Active, QPalette.ButtonText, brush)
        palette3.setBrush(QPalette.Inactive, QPalette.WindowText, brush1)
        palette3.setBrush(QPalette.Inactive, QPalette.Text, brush1)
        palette3.setBrush(QPalette.Inactive, QPalette.ButtonText, brush1)
        palette3.setBrush(QPalette.Disabled, QPalette.WindowText, brush2)
        palette3.setBrush(QPalette.Disabled, QPalette.Text, brush2)
        palette3.setBrush(QPalette.Disabled, QPalette.ButtonText, brush2)
        self.label_16.setPalette(palette3)
        self.label_16.setFont(font)

        self.gridLayout_2.addWidget(self.label_16, 0, 0, 1, 1)

        self.activePowerImbalanceSpinBox = QDoubleSpinBox(self.frame_2)
        self.activePowerImbalanceSpinBox.setObjectName(u"activePowerImbalanceSpinBox")
        self.activePowerImbalanceSpinBox.setDecimals(1)
        self.activePowerImbalanceSpinBox.setMaximum(999.000000000000000)
        self.activePowerImbalanceSpinBox.setValue(3.000000000000000)

        self.gridLayout_2.addWidget(self.activePowerImbalanceSpinBox, 1, 1, 1, 1)


        self.horizontalLayout.addWidget(self.frame_2)

        self.horizontalSpacer = QSpacerItem(565, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer)

        icon3 = QIcon()
        icon3.addFile(u":/icons/icons/gear.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.tabWidget.addTab(self.tab_2, icon3, "")

        self.verticalLayout_2.addWidget(self.tabWidget)

        MainWindow.setCentralWidget(self.centralwidget)
        self.statusbar = QStatusBar(MainWindow)
        self.statusbar.setObjectName(u"statusbar")
        MainWindow.setStatusBar(self.statusbar)
        self.toolBar = QToolBar(MainWindow)
        self.toolBar.setObjectName(u"toolBar")
        self.toolBar.setMovable(False)
        MainWindow.addToolBar(Qt.ToolBarArea.TopToolBarArea, self.toolBar)

        self.toolBar.addAction(self.actionAnalyze)
        self.toolBar.addAction(self.actionFix_issues)
        self.toolBar.addAction(self.actionSave_diagnostic)

        self.retranslateUi(MainWindow)

        self.tabWidget.setCurrentIndex(0)


        QMetaObject.connectSlotsByName(MainWindow)
    # setupUi

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle(QCoreApplication.translate("MainWindow", u"Grid Analysis Dialog", None))
        self.actionSave_diagnostic.setText(QCoreApplication.translate("MainWindow", u"Save diagnostic", None))
        self.actionFix_issues.setText(QCoreApplication.translate("MainWindow", u"Fix issues", None))
        self.actionAnalyze.setText(QCoreApplication.translate("MainWindow", u"Analyze", None))
#if QT_CONFIG(tooltip)
        self.actionAnalyze.setToolTip(QCoreApplication.translate("MainWindow", u"Analyze the grid data", None))
#endif // QT_CONFIG(tooltip)
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab), QCoreApplication.translate("MainWindow", u"Analysis", None))
        self.label_4.setText(QCoreApplication.translate("MainWindow", u"Transformers", None))
#if QT_CONFIG(tooltip)
        self.label_5.setToolTip(QCoreApplication.translate("MainWindow", u"VCC: Short circuit voltage from the short circuit study", None))
#endif // QT_CONFIG(tooltip)
        self.label_5.setText(QCoreApplication.translate("MainWindow", u"VCC ", None))
        self.transformerVccMaxSpinBox.setSuffix(QCoreApplication.translate("MainWindow", u" %", None))
        self.label.setText(QCoreApplication.translate("MainWindow", u"Voltage set point", None))
#if QT_CONFIG(tooltip)
        self.virtualTapToleranceSpinBox.setToolTip(QCoreApplication.translate("MainWindow", u"This is the tolerance for the difference between a transformer winding nominal voltage and the nominal voltage of the bus it is connected to.", None))
#endif // QT_CONFIG(tooltip)
        self.virtualTapToleranceSpinBox.setSuffix(QCoreApplication.translate("MainWindow", u" %", None))
        self.label_3.setText(QCoreApplication.translate("MainWindow", u"Tap angles", None))
        self.label_12.setText(QCoreApplication.translate("MainWindow", u"Generators", None))
        self.label_11.setText("")
        self.label_2.setText(QCoreApplication.translate("MainWindow", u"Tap module", None))
#if QT_CONFIG(tooltip)
        self.label_15.setToolTip(QCoreApplication.translate("MainWindow", u"Percentual difference between the nominal voltages at both sides of a line", None))
#endif // QT_CONFIG(tooltip)
        self.label_15.setText(QCoreApplication.translate("MainWindow", u"Vnom difference", None))
#if QT_CONFIG(tooltip)
        self.lineNominalVoltageToleranceSpinBox.setToolTip(QCoreApplication.translate("MainWindow", u"Percentual difference between the nominal voltages at both sides of a line", None))
#endif // QT_CONFIG(tooltip)
        self.lineNominalVoltageToleranceSpinBox.setSuffix(QCoreApplication.translate("MainWindow", u" %", None))
        self.label_13.setText("")
        self.label_14.setText(QCoreApplication.translate("MainWindow", u"Lines", None))
        self.label_6.setText(QCoreApplication.translate("MainWindow", u"Virtual tap tolerance", None))
        self.transformerVccMinSpinBox.setSuffix(QCoreApplication.translate("MainWindow", u" %", None))
        self.fixTimeSeriesCheckBox.setText(QCoreApplication.translate("MainWindow", u"Analyze / Fix time series values", None))
        self.label_8.setText(QCoreApplication.translate("MainWindow", u"Active power imbalance", None))
        self.label_16.setText(QCoreApplication.translate("MainWindow", u"System", None))
        self.activePowerImbalanceSpinBox.setSuffix(QCoreApplication.translate("MainWindow", u" %", None))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_2), QCoreApplication.translate("MainWindow", u"Settings", None))
        self.toolBar.setWindowTitle(QCoreApplication.translate("MainWindow", u"toolBar", None))
    # retranslateUi




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'SelectComponents.ui'
##
## Created by: Qt User Interface Compiler version 6.7.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractButton, QApplication, QCheckBox, QDialog,
    QDialogButtonBox, QFrame, QGridLayout, QLabel,
    QSizePolicy, QVBoxLayout, QWidget)
from .icons_rc import *
from .icons_rc import *

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        if not MainWindow.objectName():
            MainWindow.setObjectName(u"MainWindow")
        MainWindow.resize(527, 368)
        self.verticalLayout = QVBoxLayout(MainWindow)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout_2 = QVBoxLayout()
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.frame = QFrame(MainWindow)
        self.frame.setObjectName(u"frame")
        self.frame.setFrameShape(QFrame.StyledPanel)
        self.frame.setFrameShadow(QFrame.Raised)
        self.verticalLayout_3 = QVBoxLayout(self.frame)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.gridLayout = QGridLayout()
        self.gridLayout.setObjectName(u"gridLayout")
        self.checkBox = QCheckBox(self.frame)
        self.checkBox.setObjectName(u"checkBox")
        self.checkBox.setAutoExclusive(True)

        self.gridLayout.addWidget(self.checkBox, 3, 1, 1, 1)

        self.checkBox_3 = QCheckBox(self.frame)
        self.checkBox_3.setObjectName(u"checkBox_3")
        self.checkBox_3.setAutoExclusive(True)

        self.gridLayout.addWidget(self.checkBox_3, 5, 1, 1, 1)

        self.checkBox_7 = QCheckBox(self.frame)
        self.checkBox_7.setObjectName(u"checkBox_7")
        self.checkBox_7.setAutoExclusive(True)

        self.gridLayout.addWidget(self.checkBox_7, 9, 1, 1, 1)

        self.checkBox_2 = QCheckBox(self.frame)
        self.checkBox_2.setObjectName(u"checkBox_2")
        self.checkBox_2.setAutoExclusive(True)

        self.gridLayout.addWidget(self.checkBox_2, 4, 1, 1, 1)

        self.checkBox_4 = QCheckBox(self.frame)
        self.checkBox_4.setObjectName(u"checkBox_4")
        self.checkBox_4.setAutoExclusive(True)

        self.gridLayout.addWidget(self.checkBox_4, 6, 1, 1, 1)

        self.label_2 = QLabel(self.frame)
        self.label_2.setObjectName(u"label_2")

        self.gridLayout.addWidget(self.label_2, 10, 1, 1, 1)

        self.label = QLabel(self.frame)
        self.label.setObjectName(u"label")
        font = QFont()
        font.setPointSize(16)
        self.label.setFont(font)

        self.gridLayout.addWidget(self.label, 1, 1, 1, 1)

        self.checkBox_5 = QCheckBox(self.frame)
        self.checkBox_5.setObjectName(u"checkBox_5")
        self.checkBox_5.setAutoExclusive(True)

        self.gridLayout.addWidget(self.checkBox_5, 7, 1, 1, 1)

        self.checkBox_6 = QCheckBox(self.frame)
        self.checkBox_6.setObjectName(u"checkBox_6")
        self.checkBox_6.setAutoExclusive(True)

        self.gridLayout.addWidget(self.checkBox_6, 8, 1, 1, 1)

        self.label_3 = QLabel(self.frame)
        self.label_3.setObjectName(u"label_3")

        self.gridLayout.addWidget(self.label_3, 2, 1, 1, 1)


        self.verticalLayout_3.addLayout(self.gridLayout)

        self.buttonBox = QDialogButtonBox(self.frame)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Ok)

        self.verticalLayout_3.addWidget(self.buttonBox)


        self.verticalLayout_2.addWidget(self.frame)


        self.verticalLayout.addLayout(self.verticalLayout_2)


        self.retranslateUi(MainWindow)

        QMetaObject.connectSlotsByName(MainWindow)
    # setupUi

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle(QCoreApplication.translate("MainWindow", u"Dialog", None))
        self.checkBox.setText(QCoreApplication.translate("MainWindow", u"Transformer", None))
        self.checkBox_3.setText(QCoreApplication.translate("MainWindow", u"Line", None))
        self.checkBox_7.setText(QCoreApplication.translate("MainWindow", u"Substation", None))
        self.checkBox_2.setText(QCoreApplication.translate("MainWindow", u"Cable", None))
        self.checkBox_4.setText(QCoreApplication.translate("MainWindow", u"Turbine", None))
        self.label_2.setText(QCoreApplication.translate("MainWindow", u"(Only one may be selected at a time.)", None))
        self.label.setText(QCoreApplication.translate("MainWindow", u"Insert Catalogue Component", None))
        self.checkBox_5.setText(QCoreApplication.translate("MainWindow", u"Reactor", None))
        self.checkBox_6.setText(QCoreApplication.translate("MainWindow", u"Switchgear", None))
        self.label_3.setText(QCoreApplication.translate("MainWindow", u"Please select the component data to add to the catalogue:", None))
    # retranslateUi




# MIT License
#
# Copyright (c) 2021-2022 Yunosuke Ohsugi
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module setting up ui of mdi window."""
from VeraGrid.ThirdParty.qdarktheme.qtpy.QtWidgets import (
    QHBoxLayout,
    QLabel,
    QMdiArea,
    QMdiSubWindow,
    QPushButton,
    QSplitter,
    QTextEdit,
    QVBoxLayout,
    QWidget,
)


class MdiUI:
    """The ui class of mdi window."""

    def _make_mdi_area_test_widget(self, enable_tab_mode=False):
        # Widgets
        container = QWidget()
        mdi_area = QMdiArea()
        label_test_name = QLabel()
        cascade_button = QPushButton("Cascade")
        new_button = QPushButton("Add new")
        tiled_button = QPushButton("Tiled")

        # Setup widgets
        if enable_tab_mode:
            mdi_area.setViewMode(QMdiArea.ViewMode.TabbedView)
            label_test_name.setText("QMdiArea(QMdiArea.viewMode = TabbedView)")
        else:
            label_test_name.setText("QMdiArea(QMdiArea.viewMode = SubWindowView)")

        def add_new_sub_window():
            sub_win = QMdiSubWindow(container)
            sub_win_main_widget = QWidget(sub_win)
            v_layout = QVBoxLayout(sub_win_main_widget)
            v_layout.addWidget(QTextEdit("Sub window"))

            sub_win.setWidget(sub_win_main_widget)
            mdi_area.addSubWindow(sub_win)
            sub_win.show()

        add_new_sub_window()
        new_button.pressed.connect(add_new_sub_window)
        cascade_button.pressed.connect(mdi_area.cascadeSubWindows)
        tiled_button.pressed.connect(mdi_area.tileSubWindows)
        new_button.setDefault(True)

        # Layout
        h_layout = QHBoxLayout()
        h_layout.addWidget(new_button)
        h_layout.addWidget(cascade_button)
        h_layout.addWidget(tiled_button)

        v_main_layout = QVBoxLayout(container)
        v_main_layout.addWidget(label_test_name)
        v_main_layout.addLayout(h_layout)
        v_main_layout.addWidget(mdi_area)
        return container

    def setup_ui(self, win: QWidget) -> None:
        """Set up ui."""
        # Widgets
        splitter = QSplitter()

        # Setup widgets
        mdi_area = self._make_mdi_area_test_widget()
        mdi_area_with_tab = self._make_mdi_area_test_widget(enable_tab_mode=True)

        # Layout
        splitter.addWidget(mdi_area)
        splitter.addWidget(mdi_area_with_tab)

        main_layout = QVBoxLayout(win)
        main_layout.addWidget(splitter)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

import sys
from PySide6 import QtWidgets
from PySide6.QtCore import Qt
from VeraGrid.Gui.SyncDialogue.gui import Ui_Dialog
from VeraGrid.Session.synchronization_driver import get_issues_tree_view_model, FileSyncThread


class SyncDialogueWindow(QtWidgets.QDialog):
    """
    SyncDialogueWindow
    """
    def __init__(self, file_sync_thread: FileSyncThread, parent=None):
        """

        :param parent:
        """
        QtWidgets.QDialog.__init__(self, parent)
        self.ui = Ui_Dialog()
        self.ui.setupUi(self)
        self.setWindowTitle('Sync conflicts')

        self.file_sync_thread = file_sync_thread

        self.file_sync_thread.pause()

        self.model = get_issues_tree_view_model(self.file_sync_thread.issues)
        self.ui.treeView.setModel(self.model)
        self.ui.treeView.expandAll()

        self.ui.accept_selected_pushButton.clicked.connect(self.accept_changes)
        self.ui.reject_selected_pushButton.clicked.connect(self.reject_changes)
        self.ui.doit_pushButton.clicked.connect(self.doit)

    def msg(self, text, title="Warning"):
        """
        Message box
        :param text: Text to display
        :param title: Name of the window
        """
        msg = QtWidgets.QMessageBox()
        msg.setIcon(QtWidgets.QMessageBox.Icon.Information)
        msg.setText(text)
        # msg.setInformativeText("This is additional information")
        msg.setWindowTitle(title)
        # msg.setDetailedText("The details are as follows:")
        msg.setStandardButtons(QtWidgets.QMessageBox.StandardButton.Ok)
        retval = msg.exec()

    def closeEvent(self, event):
        self.file_sync_thread.resume()

    def get_item_level(self, item):

        l = 0
        idx = item
        while idx._parent().data() is not None:
            idx = idx._parent()
            l += 1
        return l

    def set_selection_value(self, value=True):
        """
        Change the selected tree items
        :param value: True / False
        :return: Nothing
        """
        idx = self.ui.treeView.selectedIndexes()

        # filter by indices of level 2
        for i in idx:

            c = i.column()
            r = i.row()

            if c == 0 or c == 5:
                if self.get_item_level(i) == 2:

                    if c == 0:  # if it is the first column process the issue index
                        issue_idx = int(i.data())
                        self.file_sync_thread.issues[issue_idx].__accept__ = value

                    elif c == 5:   # if it is the status column, process the status

                        item = self.ui.treeView.model().itemFromIndex(i)
                        if value:
                            item.setCheckState(Qt.Checked)
                        else:
                            item.setCheckState(Qt.Unchecked)

    def accept_changes(self):
        """
        Accept selected issues
        """
        self.set_selection_value(True)

    def reject_changes(self):
        """
        Reject selected issues
        """
        self.set_selection_value(False)

    def doit(self):
        """
        process issues and close
        """

        self.file_sync_thread.process_issues()
        self.close()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
import io
import numpy as np
import pandas as pd
from PySide6 import QtCore, QtWidgets

from VeraGrid.Gui.messages import error_msg
from VeraGrid.Gui.wrappable_table_model import WrappableTableModel
from VeraGridEngine.Simulations.results_table import ResultsTable
from VeraGridEngine.Utils.Filtering.results_table_filtering import FilterResultsTable


def fast_data_to_numpy_text(data: np.ndarray) -> str:
    """
    Convert numpy array to text (as to copy/paste in a python console)
    :param data: numpy array
    :return: numpy declarative string
    """
    if len(data.shape) == 1:
        txt = '[' + ', '.join(['{0:.6f}'.format(x) for x in data]) + ']'

    elif len(data.shape) == 2:

        if data.shape[1] > 1:
            # header first
            txt = '['

            # data
            for t in range(data.shape[0]):
                txt += '[' + ', '.join(['{0:.6f}'.format(x) for x in data[t, :]]) + '],\n'

            txt += ']'
        else:
            txt = '[' + ', '.join(['{0:.6f}'.format(x) for x in data[:, 0]]) + ']'
    else:
        txt = '[]'

    return txt


class ResultsModel(WrappableTableModel):
    """
    Class to populate a Qt table view with data from the results
    """
    def __init__(self, table: ResultsTable, parent=None):
        """

        :param table:
        """
        WrappableTableModel.__init__(self, parent)

        self.table = table

        self.units = table.units

        self.max_to_min = np.ones(self.table.c, dtype=bool)

    def flags(self, index: QtCore.QModelIndex) -> QtCore.Qt.ItemFlag:
        """

        :param index:
        :return:
        """
        if self.table.editable and index.column() > self.table.editable_min_idx:
            return QtCore.Qt.ItemFlag.ItemIsEditable | QtCore.Qt.ItemFlag.ItemIsEnabled | QtCore.Qt.ItemFlag.ItemIsSelectable
        else:
            return QtCore.Qt.ItemFlag.ItemIsEnabled | QtCore.Qt.ItemFlag.ItemIsSelectable

    def update(self) -> None:
        """
        update table
        """
        self.layoutAboutToBeChanged.emit()
        self.layoutChanged.emit()

    def rowCount(self, parent: QtCore.QModelIndex = None) -> int:
        """

        :param parent:
        :return:
        """
        return self.table.r

    def columnCount(self, parent: QtCore.QModelIndex = None) -> int:
        """

        :param parent:
        :return:
        """
        return self.table.c

    def data(self, index: QtCore.QModelIndex, role=QtCore.Qt.ItemDataRole.DisplayRole):
        """

        :param index:
        :param role:
        :return:
        """
        if index.isValid():

            val = self.table.data_c[index.row(), index.column()]

            if role == QtCore.Qt.ItemDataRole.DisplayRole:

                if isinstance(val, str):
                    return val
                elif isinstance(val, complex):
                    if val.real != 0 or val.imag != 0:
                        return val.__format__(self.table.format_string)
                    else:
                        return '0'
                else:
                    if val != 0:
                        return val.__format__(self.table.format_string)
                    else:
                        return '0'

            elif role == QtCore.Qt.ItemDataRole.BackgroundRole:

                return None  # QBrush(Qt.yellow)

        return None

    def headerData(self,
                   section: int,
                   orientation: QtCore.Qt.Orientation,
                   role=QtCore.Qt.ItemDataRole.DisplayRole):
        """
        Get the header value
        :param section: header index
        :param orientation: Orientation {QtCore.Qt.Horizontal, QtCore.Qt.Vertical}
        :param role:
        :return:
        """
        if self._hide_headers_mode is True:
            return None

        if role == QtCore.Qt.ItemDataRole.DisplayRole:
            if orientation == QtCore.Qt.Orientation.Horizontal:
                if len(self.table.cols_c) > section:
                    val = self.table.cols_c[section]
                else:
                    return ""

            elif orientation == QtCore.Qt.Orientation.Vertical:
                if self.table.index_c is None:
                    return section
                else:
                    val = f"{section}: {self.table.index_c[section]}"
            else:
                return ""

            if isinstance(val, pd.Timestamp):
                return val.strftime('%Y/%m/%d  %H:%M:%S')
            else:
                return val
        return None

    def slice_cols(self, col_idx) -> "ResultsModel":
        """
        Make column slicing
        :param col_idx: indices of the columns
        :return: Nothing
        """
        return ResultsModel(self.table.slice_cols(col_idx))

    def search_in_columns(self, txt):
        """
        Search stuff
        :param txt:
        :return:
        """
        print('Searching', txt)
        mdl = self.table.search_in_columns(txt)

        if mdl is not None:
            return ResultsModel(mdl)
        else:
            return None

    def transpose(self):
        """
        Transpose the results in-place
        """
        self.table.transpose()

    def search(self, txt: str):
        """
        Search stuff
        :param txt:
        :return:
        """
        filter_ = FilterResultsTable(self.table)
        filter_.parse(expression=txt)
        mdl = filter_.apply()

        if mdl is not None:
            return ResultsModel(mdl)
        else:
            return None

    def sort_column(self, c: int):
        """
        Sort column, initially max to min, but fliping the side each time
        :param c:
        :return:
        """

        self.table.sort_column(c=c, max_to_min=self.max_to_min[c])
        self.max_to_min[c] = not self.max_to_min[c]

    def copy_to_column(self, row, col):
        """
        Copies one value to all the column
        @param row: Row of the value
        @param col: Column of the value
        @return: Nothing
        """
        self.table.copy_to_column(row, col)

    def is_complex(self):
        return self.table.is_complex()

    def get_data(self):
        """
        Returns: index, columns, data
        """
        return self.table.get_data()

    def convert_to_cdf(self):
        """
        Convert the data in-place to CDF based
        :return:
        """

        # calculate the proportional values of samples
        self.table.convert_to_cdf()

    def convert_to_abs(self):
        """
        Convert the data to abs
        :return:
        """
        self.table.convert_to_abs()

    def to_df(self):
        """
        get DataFrame
        """
        return self.table.to_df()

    def save_to_excel(self, file_name):
        """
        save data to excel
        :param file_name:
        """
        self.to_df().to_excel(file_name)

    def save_to_csv(self, file_name):
        """
        Save data to csv
        :param file_name:
        """
        self.to_df().to_csv(file_name)

    def get_data_frame(self):
        """
        Save data to csv
        """
        return self.table.get_data_frame()

    def copy_to_clipboard(self):
        """
        Copy profiles to clipboard
        """
        n = len(self.table.cols_c)

        if n > 0:

            df = self.get_data_frame()
            s = io.StringIO()
            df.to_csv(s, sep='\t')
            txt = s.getvalue()

            # copy to clipboard
            cb = QtWidgets.QApplication.clipboard()
            cb.clear()
            cb.setText(txt)

        else:
            # there are no elements
            pass

    def copy_numpy_to_clipboard(self):
        """
        Copy profiles to clipboard
        """
        n = len(self.table.cols_c)

        if n > 0:

            index, columns, data = self.get_data()

            txt = fast_data_to_numpy_text(data)

            # copy to clipboard
            cb = QtWidgets.QApplication.clipboard()
            cb.clear()
            cb.setText(txt)

        else:
            # there are no elements
            pass

    def plot(self, ax=None, selected_col_idx=None, selected_rows=None, stacked=False):
        """
        Plot the data model
        :param ax: Matplotlib axis
        :param selected_col_idx: list of selected column indices
        :param selected_rows: list of rows to plot
        :param stacked: stack the data?
        """
        try:
            self.table.plot(ax=ax,
                            selected_col_idx=selected_col_idx,
                            selected_rows=selected_rows,
                            stacked=stacked)
        except ValueError as e:
            error_msg(text=str(e), title="Plotting error")
